import keyword as _keyword
import re as _re
from numbers import Number
import numpy
import copy

from .util.naming import parenthize, valid_identifier_or_parenthized_string
from .util.touch_notifier import TouchNotify

_chi = u"\u03C7"
_beta = u"\u03B2"
_hahook = u"\u04FC"

_ParameterRef_repr_txt = 'P'
_DataRef_repr_txt = 'X'


class Role(type):
	def __getattr__(cls, key):
		try:
			return super().getattr(key)
		except AttributeError:
			if cls is ParameterRef or cls is DataRef or cls is CombinedRef:
				return cls(key)
			else:
				raise


_data_description_catalog = {}


class DataRef(str, metaclass=Role):
	def __new__(cls, name, **kwarg):
		if hasattr(str, name):
			raise NameError("cannot create DataRef with the name of a str method ({})".format(name))
		if name in ('_descrip', 'descrip', '_role'):
			raise NameError("cannot create DataRef with the name ({})".format(name))
		try:
			raise TypeError("debugging")  # The getattr here is not really needed?
			return getattr(cls, name)
		except:
			self = super().__new__(cls, name)
			self._role = 'data'
			return self

	def __getnewargs__(self):
		return (str(self), )

	def __getstate__(self):
		return self.__dict__

	def __setstate__(self, d):
		self.__dict__.update(d)

	def __init__(self, name, descrip=None):
		if descrip is not None:
			_data_description_catalog[self] = descrip
		super().__init__()

	@property
	def descrip(self):
		if self in _data_description_catalog:
			return _data_description_catalog[self]
		return None

	@descrip.setter
	def descrip(self, value):
		if value is not None:
			_data_description_catalog[self] = value

	def __repr__(self):
		s = super().__str__()
		if _re.match("[_A-Za-z][_a-zA-Z0-9]*$", s) and not _keyword.iskeyword(s):
			return "{}.{}".format(_DataRef_repr_txt, s)
		else:
			return "{}('{}')".format(_DataRef_repr_txt, s)

	def __add__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			return LinearComponent(data=str(self), param="CONSTANT") + other
		return DataRef("{}+{}".format(parenthize(self), parenthize(other, True)))

	def __radd__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			return other + LinearComponent(data=str(self), param="CONSTANT")
		return DataRef("{}+{}".format(parenthize(other), parenthize(self, True)))

	def __sub__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			return LinearComponent(data=str(self), param="CONSTANT") - other
		return DataRef("{}-{}".format(parenthize(self), parenthize(other, True)))

	def __rsub__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			return other - LinearComponent(data=str(self), param="CONSTANT")
		return DataRef("{}-{}".format(parenthize(other), parenthize(self, True)))

	def __mul__(self, other):
		if isinstance(other, LinearFunction):
			trial = LinearFunction()
			for component in other:
				trial += self * component
			return trial
		if isinstance(other, LinearComponent):
			return LinearComponent(data=str(self * other.data), param=str(other.param))
		if type(other) is ParameterRef:
			return LinearComponent(data=str(self), param=str(other))
		if type(other) is _param_multiply:
			if isinstance(other._left, ParameterRef):
				return LinearComponent(data=str(self), param=str(other._left)) * other._right
			else:
				return LinearComponent(data=str(self), param=str(other._right)) * other._left
		if type(other) is _param_negate:
			return LinearComponent(data=str(-self), param=str(other._orig))
		if other == 0:
			return 0
		return DataRef("{}*{}".format(parenthize(self), parenthize(other, True)))

	def __rmul__(self, other):
		if isinstance(other, LinearFunction):
			raise NotImplementedError
		if isinstance(other, LinearComponent):
			return LinearComponent(data=str(other.data * self), param=str(other.param))
		if type(other) is ParameterRef:
			return LinearComponent(data=str(self), param=str(other))
		if type(other) is _param_negate:
			return LinearComponent(data=str(-self), param=str(other._orig))
		if other == 0:
			return 0
		return DataRef("{}*{}".format(parenthize(other), parenthize(self, True)))

	def __truediv__(self, other):
		if isinstance(other, LinearFunction):
			raise NotImplementedError
		if isinstance(other, LinearComponent):
			raise NotImplementedError
		if isinstance(other, ParameterRef):
			raise NotImplementedError
		return DataRef("{}/{}".format(parenthize(self), parenthize(other, True)))

	def __rtruediv__(self, other):
		if isinstance(other, LinearFunction):
			raise NotImplementedError
		if isinstance(other, LinearComponent):
			raise NotImplementedError
		if isinstance(other, ParameterRef):
			raise NotImplementedError
		return DataRef("{}/{}".format(parenthize(other), parenthize(self, True)))

	def __and__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			raise NotImplementedError
		return DataRef("{}&{}".format(parenthize(self), parenthize(other, True)))

	def __rand__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			raise NotImplementedError
		return DataRef("{}&{}".format(parenthize(other), parenthize(self, True)))

	def __or__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			raise NotImplementedError
		return DataRef("{}|{}".format(parenthize(self), parenthize(other, True)))

	def __ror__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			raise NotImplementedError
		return DataRef("{}|{}".format(parenthize(other), parenthize(self, True)))

	def __xor__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			raise NotImplementedError
		return DataRef("{}^{}".format(parenthize(self), parenthize(other, True)))

	def __rxor__(self, other):
		if isinstance(other, (ParameterRef, LinearComponent, LinearFunction)):
			raise NotImplementedError
		return DataRef("{}^{}".format(parenthize(other), parenthize(self, True)))

	def __neg__(self):
		return DataRef("-({})".format(self, True))

	def __pos__(self):
		return self

	def __eq__(self, other):
		if isinstance(other, DataRef):
			if repr(other) == repr(self):
				return True
		if isinstance(other, ParameterRef):
			return False
		if isinstance(other, str):
			if other == super().__str__():
				return True
		return False

	def __hash__(self):
		return hash(super().__str__())

	def eval(self, namespace=None, *, globals=None, **more_namespace):
		import numpy
		from .util.common_functions import piece
		use_namespace = {'exp': numpy.exp, 'log': numpy.log, 'log1p': numpy.log1p, 'fabs': numpy.fabs,
		                 'sqrt': numpy.sqrt,
		                 'absolute': numpy.absolute, 'isnan': numpy.isnan, 'isfinite': numpy.isfinite,
		                 'logaddexp': numpy.logaddexp, 'fmin': numpy.fmin, 'fmax': numpy.fmax,
		                 'nan_to_num': numpy.nan_to_num,
						 'piece': piece,}
		if namespace is not None:
			use_namespace.update(namespace)
		use_namespace.update(more_namespace)
		return eval(self, globals, use_namespace)

	def __getattr__(self, item):
		return DataRef(".".join([self,item]))

	def __copy__(self):
		# str are immutable
		return self

	def __deepcopy__(self, memodict):
		return copy.deepcopy(str(self), memodict)

class ParameterRef(str, metaclass=Role):
	"""
	An abstract reference to a parameter, which may or may not be included in any given model.

	Parameters
	----------
	name : str
		The name of the parameter to reference.
	default : numeric or None
		When a targeted model does not include the referenced parameter, use
		this value for :meth:`value` and :meth:`str`. If given as None, those
		methods will raise an exception.
	fmt : str or None
		The format string to use for :meth:`str`.
	"""

	def __new__(cls, name, default=None, fmt=None, default_value=None, format=None, scale=None):
		if fmt is None and format is not None:
			fmt = format
		if fmt is None:
			fmt = "{}"
		if default is None and default_value is not None:
			default = default_value
		self = super().__new__(cls, name)
		self._fmt = fmt
		self._name = name
		self._role = 'parameter'
		self._default_value = default
		return self

	def __repr__(self):
		s = super().__str__()
		if _re.match("[_A-Za-z][_a-zA-Z0-9]*$", s) and not _keyword.iskeyword(s):
			return "{}.{}".format(_ParameterRef_repr_txt, s)
		else:
			return "{}('{}')".format(_ParameterRef_repr_txt, s)

	def __str__(self):
		return super().__str__()

	def __xml__(self, resolve_parameters=None):
		from xmle import Elem
		x = Elem('div')
		a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self))
		if resolve_parameters is not None:
			plabel = "{0:.3g}".format(resolve_parameters.pvalue(self, default_value="This is a Parameter"))
		else:
			plabel = "This is a Parameter"
		a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
		return x

	def value(self, m):
		"""
		The value of the parameter in a given model.

		Parameters
		----------
		m : Model
			The model from which to extract a parameter value.

		Raises
		------
		LarchError
			When the model does not contain a parameter with the same
			name as this ParameterRef, and the default_value for this
			ParameterRef is None.
		"""
		try:
			v = m.pvalue(str(self))
		except KeyError:
			if self._default_value is not None:
				v = self._default_value
			else:
				raise
		return v

	def name(self, n):
		"""
		Set or re-set the name of the parameter referenced.

		Parameters
		----------
		n : str
			The new name of the parameter.

		Returns
		-------
		ParameterRef
			This method returns the self object, to facilitate method chaining.
		"""
		self._name = n
		return self

	def fmt(self, format):
		"""
		Set or re-set the format string for the parameter referenced.

		Parameters
		----------
		format : str
			The new format string. This will be used to format the parameter value when calling :meth:`str`.

		Returns
		-------
		ParameterRef
			This method returns the self object, to facilitate method chaining.
		"""
		if format is None or format == "":
			self._fmt = "{}"
		else:
			self._fmt = format
		try:
			self._fmt.format(0)
		except Exception as err:
			raise ValueError("invalid formating string") from err
		return self

	def default_value(self, val):
		"""
		Set or re-set the default value.

		Parameters
		----------
		val : numeric or None
			The new default value. This will be used by :meth:`value` when
			the parameter is not included in the targeted model. Set to None
			to raise an exception instead.

		Returns
		-------
		ParameterRef
			This method returns the self object, to facilitate method chaining.
		"""
		self._default_value = val
		return self

	def strf(self, m, fmt=None):
		"""
		Gives the :meth:`value` of the parameter in a given model as a string.

		The string is formated using the :meth:`fmt` string if given. If not
		given, Python's default string formatting is used.

		Parameters
		----------
		m : Model
			The model from which to extract a parameter value.
		fmt : str or None
			If not None, a format string which will be used, overriding any
			previous setting by a :meth:`fmt` command.  The override is
			valid for this method call only, and does not change the
			format for future calls.

		Raises
		------
		KeyError
			When the model does not contain a parameter with the same
			name as this ParameterRef, and the default_value for this
			ParameterRef is None.
		"""
		try:
			if fmt is not None:
				return fmt.format(self.value(m))
			else:
				return self._fmt.format(self.value(m))
		except KeyError:
			return "NA"

	def getname(self):
		return self._name

	def valid(self, m):
		"""
		Check if this ParameterRef would give a value for a given model.

		Parameters
		----------
		m : Model
			The model from which to extract a parameter value.

		Returns
		-------
		bool
			False if the value method would raise an exception, and True otherwise.
		"""
		if str(self) in m:
			return True
		if self._default_value is not None:
			return True
		return False

	def __add__(self, other):
		if other == 0:
			return self
		if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
			return LinearComponent(param=str(self), data="1") + other
		return _param_add(self, other)

	def __radd__(self, other):
		if other == 0:
			return self
		if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
			return other + LinearComponent(param=str(self), data="1")
		return _param_add(other, self)

	def __sub__(self, other):
		if other == 0:
			return self
		if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
			return LinearComponent(param=str(self), data="1") - other
		return _param_subtract(self, other)

	def __rsub__(self, other):
		if other == 0:
			return self
		if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
			return other - LinearComponent(param=str(self), data="1")
		return _param_subtract(other, self)

	def __mul__(self, other):
		if isinstance(other, (LinearComponent, LinearFunction)):
			raise TypeError("cannot do ParameterRef * LinearComponent")
		if type(other) is DataRef:
			if isinstance(self, _param_multiply):
				if isinstance(self._left, ParameterRef):
					return LinearComponent(data=str(other), param=str(self._left)) * self._right
				else:
					return LinearComponent(data=str(other), param=str(self._right)) * self._left
			return LinearComponent(data=str(other), param=str(self))
		return _param_multiply(self, other)

	def __rmul__(self, other):
		if isinstance(other, (LinearComponent, LinearFunction)):
			raise TypeError("cannot do LinearComponent * ParameterRef")
		if type(other) is DataRef:
			if isinstance(self, _param_multiply):
				if isinstance(self._left, ParameterRef):
					return LinearComponent(data=str(other), param=str(self._left)) * self._right
				else:
					return LinearComponent(data=str(other), param=str(self._right)) * self._left
			return LinearComponent(data=str(other), param=str(self))
		return _param_multiply(other, self)

	def __truediv__(self, other):
		if isinstance(other, DataRef):
			return self * (1 / other)
		if isinstance(other, (LinearComponent, LinearFunction)):
			raise TypeError("cannot do ParameterRef / LinearComponent")
		return _param_divide(self, other)

	def __rtruediv__(self, other):
		if isinstance(other, DataRef):
			raise TypeError("cannot do DataRef / ParameterRef")
		if isinstance(other, (LinearComponent, LinearFunction)):
			raise TypeError("cannot do LinearComponent / ParameterRef")
		return _param_divide(other, self)

	def __neg__(self):
		return _param_negate(self)

	def __pos__(self):
		return self

	def __eq__(self, other):
		if isinstance(other, ParameterRef):
			if repr(other) == repr(self):
				return True
		if isinstance(other, DataRef):
			return False
		if isinstance(other, str):
			if other == super().__str__():
				return True
		return False

	def __hash__(self):
		return hash(repr(self))

	def _to_Linear(self):
		return LinearComponent(param=str(self), data="1")

	def _is_scaled_parameter(self):
		if type(self) is ParameterRef:
			return (self, 1.0)
		return None

	@property
	def unscaled_parameter(self):
		y = self._is_scaled_parameter()
		if y is None:
			return None
		return y[0]

	@property
	def scaling_of_parameter(self):
		y = self._is_scaled_parameter()
		if y is None:
			return 1
		return y[1]

	def __le__(self, other):
		from .model.guidance import ParameterScaledOrdering
		try:
			oth = other._is_scaled_parameter()
		except AttributeError:
			raise TypeError("other is not scaled parameter")
		if oth:
			return ParameterScaledOrdering(None, *self._is_scaled_parameter(), *oth)
		else:
			raise TypeError("other is not scaled parameter")

	def __ge__(self, other):
		from .model.guidance import ParameterScaledOrdering
		try:
			oth = other._is_scaled_parameter()
		except AttributeError:
			raise TypeError("other is not scaled parameter")
		if oth:
			return ParameterScaledOrdering(None, *oth, *self._is_scaled_parameter(), snap_out = 'greater?')
		else:
			raise TypeError("other is not scaled parameter")

	def lessthan(self, other):
		s = self._is_scaled_parameter()
		if s is None:
			raise TypeError("self is not scaled parameter")
		from .model.guidance import ParameterScaledOrdering, ParameterBound
		try:
			oth = other._is_scaled_parameter()
		except AttributeError:
			if isinstance(other, Number):
				return ParameterBound(None, *self._is_scaled_parameter(), None, other)
			raise TypeError("other is not scaled parameter")
		if oth:
			return ParameterScaledOrdering(None, *self._is_scaled_parameter(), *oth)
		else:
			raise TypeError("other is not scaled parameter")

	def __lshift__(self, other):
		return self.lessthan(other)

	def greaterthan(self, other):
		s = self._is_scaled_parameter()
		if s is None:
			raise TypeError("self is not scaled parameter")
		from .model.guidance import ParameterScaledOrdering, ParameterBound
		try:
			oth = other._is_scaled_parameter()
		except AttributeError:
			if isinstance(other, Number):
				return ParameterBound(None, None, other, *self._is_scaled_parameter())
			raise TypeError("other is not scaled parameter")
		if oth:
			return ParameterScaledOrdering(None, *oth, *self._is_scaled_parameter(), snap_out = 'greater')
		else:
			raise TypeError("other is not scaled parameter")

	def __rshift__(self, other):
		return self.greaterthan(other)

	def equalto(self, other):
		s = self._is_scaled_parameter()
		if s is None:
			raise TypeError("self is not scaled parameter")
		from .model.guidance import ParameterScaledEquality
		try:
			oth = other._is_scaled_parameter()
		except AttributeError:
			raise TypeError("other is not scaled parameter")
		if oth:
			return ParameterScaledEquality(None, *self._is_scaled_parameter(), *oth)
		else:
			raise TypeError("other is not scaled parameter")

	def __matmul__(self, other):
		return self.equalto(other)

	def __rmatmul__(self, other):
		try:
			return other.equalto(self)
		except AttributeError:
			raise NotImplementedError

	def between(self, lower, upper):
		return self.greaterthan(lower), self.lessthan(upper)

	def __copy__(self):
		# str are immutable
		return self

	def __deepcopy__(self, memodict):
		return copy.deepcopy(str(self), memodict)

class _param_math_binaryop(ParameterRef):
	def __new__(cls, left, right):
		self = super().__new__(cls, "")
		self._left = left
		self._right = right
		if isinstance(self._left, ParameterRef) and isinstance(self._right, ParameterRef):
			if left._fmt == "{}" and right._fmt != "{}":
				self._fmt = right._fmt
			else:
				self._fmt = left._fmt
		elif isinstance(self._left, ParameterRef):
			self._fmt = left._fmt
		elif isinstance(self._right, ParameterRef):
			self._fmt = right._fmt
		else:
			self._fmt = "{}"
		self._name = ""
		self._role = 'parameter_math'
		return self

	def __getnewargs__(self):
		return (self._left, self._right)

class _param_add(_param_math_binaryop):
	def value(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.value(m)
		else:
			x = self._left
		if isinstance(self._right, ParameterRef):
			x += self._right.value(m)
		else:
			x += self._right
		return x

	def valid(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.valid(m)
		else:
			x = True
		if isinstance(self._right, ParameterRef):
			x &= self._right.valid(m)
		return x

	def __repr__(self):
		return "({} + {})".format(repr(self._left), repr(self._right))

	def __str__(self):
		return "{} + {}".format(repr(self._left), repr(self._right))

	def _to_Linear(self):
		return self._left._to_Linear() + self._right._to_Linear()

	def __add__(self, other):
		if isinstance(other, (DataRef, LinearComponent, LinearFunction)):
			return self._to_Linear() + other
		return super().__add__(other)


class _param_subtract(_param_math_binaryop):
	def value(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.value(m)
		else:
			x = self._left
		if isinstance(self._right, ParameterRef):
			x -= self._right.value(m)
		else:
			x -= self._right
		return x

	def valid(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.valid(m)
		else:
			x = True
		if isinstance(self._right, ParameterRef):
			x &= self._right.valid(m)
		return x

	def __repr__(self):
		return "({} - {})".format(repr(self._left), repr(self._right))

	def __str__(self):
		return "{} - {}".format(repr(self._left), repr(self._right))


class _param_multiply(_param_math_binaryop):
	def value(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.value(m)
		else:
			x = self._left
		if isinstance(self._right, ParameterRef):
			x *= self._right.value(m)
		else:
			x *= self._right
		return x

	def valid(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.valid(m)
		else:
			x = True
		if isinstance(self._right, ParameterRef):
			x &= self._right.valid(m)
		return x

	def __repr__(self):
		return "({} * {})".format(repr(self._left), repr(self._right))

	def __str__(self):
		return "{} * {}".format(repr(self._left), repr(self._right))

	def _is_scaled_parameter(self):
		if type(self._left) is ParameterRef and isinstance(self._right, Number):
			return (self._left, self._right)
		elif type(self._right) is ParameterRef and isinstance(self._left, Number):
			return (self._right, self._left)
		else:
			return None

	def __mul__(self, other):
		x = self._is_scaled_parameter()
		if x and isinstance(other, Number):
			return x[0] * (x[1] * other)
		return super().__mul__(other)

	def __rmul__(self, other):
		x = self._is_scaled_parameter()
		if x and isinstance(other, Number):
			return x[0] * (x[1] * other)
		return super().__mul__(other)

	def __eq__(self, other):
		if isinstance(other, _param_multiply):
			if (self._left == other._left) and (self._right == other._right):
				return True
			elif (self._left == other._right) and (self._right == other._left):
				return True
			else:
				return False
		elif isinstance(other, ParameterRef):
			if (self._left == other) and (self._right == 1):
				return True
			elif (self._left == 1) and (self._right == other):
				return True
			else:
				return False
		else:
			raise NotImplementedError(f"type of other is {type(other)}")

	def __ne__(self, other):
		return not (self==other)

class _param_divide(_param_math_binaryop):
	def value(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.value(m)
		else:
			x = self._left
		try:
			if isinstance(self._right, ParameterRef):
				x /= self._right.value(m)
			else:
				x /= self._right
		except ZeroDivisionError:
			return float('NaN')
		return x

	def valid(self, m):
		if isinstance(self._left, ParameterRef):
			x = self._left.valid(m)
		else:
			x = True
		if isinstance(self._right, ParameterRef):
			x &= self._right.valid(m)
		return x

	def __repr__(self):
		return "({} / {})".format(repr(self._left), repr(self._right))

	def __str__(self):
		return "{} / {}".format(repr(self._left), repr(self._right))


class _param_negate(ParameterRef):
	def __new__(cls, orig):
		self = super().__new__(cls, "")
		self._orig = orig
		self._fmt = orig._fmt
		self._name = "-({})".format(orig.getname())
		self._role = 'parameter_math'
		return self

	def value(self, m):
		if isinstance(self._orig, ParameterRef):
			return -self._orig.value(m)
		else:
			return -self._orig

	def valid(self, m):
		if isinstance(self._orig, ParameterRef):
			return self._orig.valid(m)
		else:
			return True

	def __repr__(self):
		return "-({})".format(repr(self._orig))

	def __str__(self):
		return "-({})".format(repr(self._orig))


try:
	from numpy import log as _log, exp as _exp
except ImportError:
	from math import log as _log, exp as _exp


def log(x):
	if isinstance(x, DataRef):
		return DataRef("log({})".format(x))
	else:
		return _log(x)


def exp(x):
	if isinstance(x, DataRef):
		return DataRef("exp({})".format(x))
	else:
		return _exp(x)


class CombinedRef(metaclass=Role):
	def __new__(cls, s):
		return LinearComponent(data=DataRef(s), param=ParameterRef(s))

# +++

class LinearComponent2(tuple):
	__slots__ = []
	def __new__(cls, param, data='1', scale=1):
		return tuple.__new__(cls, (str(param), str(data), scale))

	@property
	def param(self):
		return ParameterRef(self[0])

	@property
	def data(self):
		return DataRef(self[1])

	@property
	def scale(self):
		return self[2]

	def __pos__(self):
		return self

	def __repr__(self):
		try:
			if self.scale == 1.0:
				try:
					data_is_1 = (float(self.data) == 1)
				except:
					data_is_1 = False
				if data_is_1:
					return f"({self.param!r})"
				else:
					return f"({self.param!r} * {self.data!r})"
			return f"({self.param!r} * {self.scale} * {self.data!r})"
		except AttributeError:
			return f"<{self.__class__.__name__} {id(self)} with error>"

	def __add__(self, other):
		if other == ():
			return self
		elif isinstance(other, LinearComponent2):
			return LinearFunction2([self, other])
		elif isinstance(other, LinearFunction2):
			return LinearFunction2([self, *other])
		elif other == 0:
			return self
		elif isinstance(other, ParameterRef):
			return LinearFunction2([self, LinearComponent2(param=other)])
		else:
			raise NotImplementedError()

	def __radd__(self, other):
		if other == 0:
			return self
		return other.__add__(self)

	def __mul__(self, other):
		if isinstance(other, (int, float, )):
			return self.__class__(
				param=self.param,
				data=self.data,
				scale=self.scale * other,
			)
		elif isinstance(other, (DataRef, )):
			return self.__class__(
				param=self.param,
				data=self.data * other,
				scale=self.scale,
			)
		else:
			raise TypeError(f'unsupported operand type(s) for {self.__class__.__name__}: {type(other)}')

	def __rmul__(self, other):
		if isinstance(other, (int, float, )):
			return self.__class__(
				param=self.param,
				data=self.data,
				scale=self.scale * other,
			)
		elif isinstance(other, (DataRef, )):
			return self.__class__(
				param=self.param,
				data=self.data * other,
				scale=self.scale,
			)
		else:
			raise TypeError(f'unsupported operand type(s) for {self.__class__.__name__}: {type(other)}')

	def __iter__(self):
		return iter(LinearFunction2([self]))

	def __eq__(self, other):
		if isinstance(other, LinearFunction2) and len(other) == 1:
			other = other[0]
		if not isinstance(other, LinearComponent2):
			return False
		if self.param != other.param:
			return False
		if self.data != other.data:
			return False
		if self.scale != other.scale:
			return False
		return True

	def __xml__(self, exponentiate_parameter=False, resolve_parameters=None):
		from xmle import Elem
		x = Elem('div')
		# x << tooltipped_style()

		if resolve_parameters is not None:
			if exponentiate_parameter:
				plabel = resolve_parameters.pvalue(self.param, default_value="This is a Parameter")
				if not isinstance(plabel, str):
					plabel = "exp({0:.3g})={1:.3g}".format(plabel, numpy.exp(plabel))
			else:
				plabel = "{0:.3g}".format(resolve_parameters.pvalue(self.param, default_value="This is a Parameter"))
		else:
			plabel = "This is a Parameter"


		data_tail = " * "
		try:
			if float(self.data)==1:
				data_tail = ""
		except:
			pass

		if self.scale == 1.0:
			if exponentiate_parameter:
				x.elem('span', tail="exp(")
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=")"+data_tail)
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
			else:
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=data_tail)
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
		else:
			if exponentiate_parameter:
				x.elem('span', tail="exp(")
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=f" * {self.scale}){data_tail}")
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
			else:
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=f" * {self.scale}{data_tail}")
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
		if data_tail == " * ":
			a_x = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.data))
			a_x.elem('span', attrib={'class':'tooltiptext'}, text="This is Data")
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()

	def _evaluate(self, p_getter, x_namespace=None, **kwargs):
		return self.scale * p_getter(str(self.param)) * self.data.eval(namespace=x_namespace, **kwargs)

	def __copy__(self):
		return self.__class__(
			param=self.param,
			data=self.data,
			scale=self.scale,
		)



from collections.abc import MutableSequence


def _try_mangle(instance):
	try:
		instance.mangle()
	except AttributeError as err:
		pass # print(f"No Mangle R: {err}")


class LinearFunction2(MutableSequence):

	def __init__(self, init=None, touch_callback=None):
		self.__func = list()
		self.__instance = None
		if init is not None:
			for i in init:
				if isinstance(i, LinearComponent2):
					self.__func.append(i)
				else:
					raise TypeError(f'members of {self.__class__.__name__} must be LinearComponent2')

	def set_instance(self, instance):
		self.__instance = instance

	def __get__(self, instance, owner):
		#print(f"get from descriptor object\n  instance={instance}\n  owner={owner}")
		self.__instance = instance
		return self

	def __set__(self, instance, values):
		#print(f"set in descriptor object\n  instance={instance}\n  values={values}")
		self.__init__(values)
		self.__instance = instance
		_try_mangle(self.__instance)

	def __delete__(self, instance):
		#print("deleted in descriptor object", instance)
		self.__instance = instance
		self.__func = list()
		_try_mangle(self.__instance)

	def __set_name__(self, owner, name):
		#print(f"setname in descriptor object\n  owner={owner}\n  name={name}\n  self={self}")
		self.name = name

	def __getitem__(self, item):
		return self.__func[item]

	def __setitem__(self, key, value):
		if isinstance(value, LinearComponent2):
			self.__func[key] = value
			_try_mangle(self.__instance)
		else:
			raise TypeError(f'members of {self.__class__.__name__} must be LinearComponent2')

	def __delitem__(self, key):
		del self.__func[key]
		_try_mangle(self.__instance)

	def __len__(self):
		return len(self.__func)

	def insert(self, index, value):
		if isinstance(value, LinearComponent2):
			self.__func.insert(index, value)
			_try_mangle(self.__instance)
		else:
			raise TypeError(f'members of {self.__class__.__name__} must be LinearComponent2')

	def __add__(self, other):
		if other == ():
			return self
		if isinstance(other, LinearFunction2):
			return type(self)(*self.__func, *other.__func)
		if isinstance(other, ParameterRef):
			other = LinearComponent2(param=other)
		if isinstance(other, LinearComponent2):
			result = type(self)(self)
			result.append(other)
			return result
		raise TypeError("cannot add type {} to LinearFunction".format(type(other)))

	def __iadd__(self, other):
		if isinstance(other, ParameterRef):
			other = LinearComponent2(param=other)
		if other == ():
			return self
		elif isinstance(other, LinearFunction2):
			self.__func.extend(other)
		elif isinstance(other, LinearComponent2):
			self.append(other)
		else:
			raise TypeError("cannot add type {} to LinearFunction".format(type(other)))
		return self

	def __radd__(self, other):
		return LinearFunction2([*other, *self])

	def __pos__(self):
		return self

	def __mul__(self, other):
		trial = LinearFunction2()
		for component in self:
			trial.append(component * other)
		return trial

	def __rmul__(self, other):
		trial = LinearFunction2()
		for component in self:
			trial += other * component
		return trial

	def __contains__(self, val):
		if isinstance(val, ParameterRef):
			for i in self:
				if i.param == val:
					return True
			return False
		if isinstance(val, DataRef):
			for i in self:
				if i.data == val:
					return True
			return False
		raise TypeError("the searched for content must be of type ParameterRef or DataRef")

	def _index_of(self, val):
		if isinstance(val, ParameterRef):
			for n, i in enumerate(self):
				if i.param == val:
					return n
			raise KeyError('ParameterRef not found')
		if isinstance(val, DataRef):
			for n, i in enumerate(self):
				if i.data == val:
					return n
			raise KeyError('DataRef not found')
		raise TypeError("the searched for content must be of type ParameterRef or DataRef")

	#######
	def reformat_param(self, container=None, pattern=None, repl=None, **kwargs):
		"""
		Transform all the parameters in the LinearFunction.

		Parameters
		----------
		container : str
			A format string, into which the previous parameters are formatted.
			Use this to append things to the parameter names.
		pattern : str
		repl : str
			Passed to `re.sub` with each existing parameter as the base string
			to be searched.

		Examples
		--------
		>>> from larch.roles import P,X
		>>> f = P.InVehTime * X.IVTT + P.OutOfVehTime * X.OVTT
		>>> f1 = f.reformat_param('{}_Suffix')
		>>> str(f1)
		'(P.InVehTime_Suffix * X.IVTT) + (P.OutOfVehTime_Suffix * X.OVTT)'
		>>> f2 = f.reformat_param(pattern='(Veh)', repl='Vehicle')
		>>> str(f2)
		'(P.InVehicleTime * X.IVTT) + (P.OutOfVehicleTime * X.OVTT)'

		"""
		import re
		r = self.__class__()
		for i in self:
			if pattern is None:
				param = i.param
			else:
				if repl is None:
					raise TypeError('must give repl with pattern')
				param = re.sub(pattern, repl, i.param, **kwargs)
			if container is None:
				container = '{}'
			r += LinearComponent2(data=i.data, param=container.format(param), scale=i.scale)
		try:
			r._x_ident = self._x_ident
		except AttributeError:
			pass
		return r

	def reformat_data(self, container=None, pattern=None, repl=None, **kwargs):
		"""
		Transform all the data in the LinearFunction.

		Parameters
		----------
		container : str
			A format string, into which the previous data strings are formatted.
			Use this to apply common global transforms to the data.
		pattern : str
		repl : str
			Passed to `re.sub` with each existing data string as the base string
			to be searched.
		"""
		import re
		r = self.__class__()
		for i in self:
			if pattern is None:
				data = i.data
			else:
				if repl is None:
					raise TypeError('must give repl with pattern')
				data = re.sub(pattern, repl, i.data, **kwargs)
			if container is None:
				container = '{}'
			r += LinearComponent2(data=container.format(data), param=i.param, scale=i.scale)
		try:
			r._x_ident = self._x_ident
		except AttributeError:
			pass
		return r

	def __code__(self):
		return " + ".join(f"({repr(i)})" for i in self)

	def __eq__(self, other):
		if not isinstance(other, LinearFunction2):
			return False
		if len(self) != len(other):
			return False
		for i, j in zip(self, other):
			if i != j: return False
		return True

	# def __getstate__(self):
	# 	state = {}
	# 	state['code'] = self.__code__()
	# 	try:
	# 		state['_x_ident'] = self._x_ident
	# 	except AttributeError:
	# 		pass
	# 	return state
	#
	# def __setstate__(self, state):
	# 	self.__init__()
	# 	if state['code']:
	# 		self.__iadd__(eval(state['code']))
	# 	if '_x_ident' in state:
	# 		self._x_ident = state['_x_ident']

	def __repr__(self):
		if len(self):
			result = " + ".join(repr(i) for i in self)
			if len(result)<80:
				return result
			else:
				return "  "+result.replace(" + ","\n+ ")
		return f"<Empty {self.__class__.__name__}>"

	def __xml__(self, linebreaks=False, lineprefix="", exponentiate_parameters=False, resolve_parameters=None):
		from xmle import Elem
		x = Elem('div', attrib={'class':'LinearFunc'})
		for n,i in enumerate(self):
			ix_ = list(i.__xml__(exponentiate_parameter=exponentiate_parameters, resolve_parameters=resolve_parameters))
			if linebreaks:
				if n>0 or lineprefix:
					ix_.insert(0,Elem('br', tail = lineprefix+" + "))
			else:
				if n < len(self)-1:
					if ix_[-1].tail is None:
						ix_[-1].tail = " + "
					else:
						ix_[-1].tail += " + "
			for ii in ix_:
				x << ii
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()

	def data(self, cls=None):
		if cls is None:
			return [_.data for _ in self]
		else:
			return [cls(_.data) for _ in self]

	def _evaluate(self, p_getter, x_namespace=None, **more_x_namespace):
		if hasattr(p_getter,'pvalue') and callable(p_getter.pvalue):
			p_getter = p_getter.pvalue
		return sum(j._evaluate(p_getter, x_namespace=x_namespace, **more_x_namespace) for j in self)

	def copy(self):
		result = self.__class__(self)
		return result

	def __deepcopy__(self, memodict):
		result = self.__class__()
		for i in self:
			result.append(copy.deepcopy(i, memodict))
		return result

	def _linear_plot_2d_data(self, p_getter, x_name, x_min, x_max, n_points=100, **other_namespace):
		import numpy
		if hasattr(self, 'plotting_namespace') and len(other_namespace)==0:
			other_namespace = self.plotting_namespace
		x = numpy.linspace(x_min, x_max, n_points)
		y = self._evaluate(p_getter, {x_name:x}, **other_namespace)
		return x,y

	def linear_plot_2d(self, p_getter, x_name, x_min, x_max, n_points=100, *, xlabel=None, svg=True, header=None, **other_namespace):

		# Delayed evaluation mode...
		if p_getter is None:
			return lambda x: self.linear_plot_2d(x, x_name, x_min, x_max, n_points=n_points, xlabel=xlabel, svg=svg, header=header, **other_namespace)

		# Active evaluation mode...
		from matplotlib import pyplot as plt
		plt.clf()
		x,y = self._linear_plot_2d_data(p_getter, x_name, x_min, x_max, n_points, **other_namespace)
		if hasattr(self, 'plotting_label'):
			plt.plot(x, y, label=self.plotting_label)
		else:
			plt.plot(x, y)
		if xlabel is None:
			plt.xlabel(x_name)
		else:
			plt.xlabel(xlabel)
		plt.tight_layout(pad=0.5)
		from .util.plotting import plot_as_svg_xhtml
		if svg is True:
			svg = {}
		if svg or svg == {}:
			if header is not None:
				svg['header'] = header
			return plot_as_svg_xhtml(plt, **svg)
		else:
			plt.show()

	def _inplot_linear_plot_2d(self, plt, p_getter, x_name, x_min, x_max, n_points=100, *, xlabel=None, svg=True, header=None, **other_namespace):

		# Delayed evaluation mode...
		if p_getter is None:
			return lambda x: self._inplot_linear_plot_2d(plt, x, x_name, x_min, x_max, n_points=n_points, xlabel=xlabel, svg=svg, header=header, **other_namespace)

		# Active evaluation mode...
		x,y = self._linear_plot_2d_data(p_getter, x_name, x_min, x_max, n_points, **other_namespace)
		if hasattr(self, 'plotting_label'):
			plt.plot(x, y, label=self.plotting_label)
		else:
			plt.plot(x,y)

	def total_ordering_increasing(self):
		from toolz.itertoolz import sliding_window
		snaps = []
		for windows_size in range(2, len(self)-1):
			for sub_p in sliding_window(windows_size, self):
				snaps.append(sub_p[0].param.lessthan(sub_p[-1].param))

		return snaps


from collections.abc import MutableMapping

class DictOfLinearFunction2(MutableMapping):

	def __init__(self, mapping=None, alts_validator=None, touch_callback=None, **kwargs):
		self.__map = {}
		if mapping is None:
			mapping = {}
		for k,v in mapping.items():
			self.__map[k] = LinearFunction2(v)
		for k,v in kwargs.items():
			try:
				self.__map[k] = LinearFunction2(v)
			except:
				print(v)
				print(type(v))
				raise

		self._alts_validator = alts_validator
		self.__instance = None

	def __get__(self, instance, owner):
		#print(f"get from descriptor object\n  instance={instance}\n  owner={owner}")
		self.__instance = instance
		return self

	def __set__(self, instance, values):
		#print(f"set in descriptor object\n  instance={instance}\n  values={values}")
		self.__init__(values)
		self.__instance = instance
		_try_mangle(self.__instance)

	def __delete__(self, instance):
		#print("deleted in descriptor object", instance)
		self.__instance = instance
		self.__func = list()
		_try_mangle(self.__instance)

	def __set_name__(self, owner, name):
		#print(f"setname in descriptor object\n  owner={owner}\n  name={name}\n  self={self}")
		self.name = name


	def set_alts_validator(self, av):
		self._alts_validator = av

	def __getitem__(self, k):
		try:
			v = self.__map[k]
			v.set_instance(self.__instance)
			return v
		except KeyError:
			if self._alts_validator is None or self._alts_validator(k):
				v = self.__map[k] = LinearFunction2()
				v.set_instance(self.__instance)
				return v
			else:
				raise

	def __setitem__(self, k, v):
		if isinstance(v, ParameterRef):
			v = v * DataRef('1')
		if isinstance(v, int) and v==0:
			v = LinearFunction2()
		elif isinstance(v, LinearComponent2):
			v = LinearFunction2([v])
		elif isinstance(v, list):
			v = LinearFunction2(v)
		elif not isinstance(v, LinearFunction2):
			raise TypeError(f"only accepts LinearFunction2 values, not {type(v)}")
		if "with error" in repr(v):
			raise ValueError("found error here")
		v.set_instance(self.__instance)
		self.__map[k] = v
		_try_mangle(self.__instance)

	def __delitem__(self, key):
		del self.__map[key]
		_try_mangle(self.__instance)

	def __iter__(self):
		return iter(self.__map.keys())

	def __len__(self):
		return len(self.__map)

	def copy(self):
		return type(self)(self)

	def __repr__(self):
		return '{0}({1})'.format(type(self).__name__, repr(self.__map))

	def __xml__(self):
		from pprint import pformat
		from xmle import Elem
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;")
		t.elem('caption', text=f"<larch.{self.__class__.__name__}>", style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;")
		if len(self):
			tr = t.elem('tr')
			tr.elem('th', text="alt")
			tr.elem('th', text='formula')
			for k,v in self.items():
				tr = t.elem('tr')
				tr.elem('td', text=str(k))
				try:
					v_ = v.__xml__()
				except AttributeError:
					tr.elem('td', text=str(v))
				else:
					tr.elem('td') << v_
		else:
			tr = t.elem('tr')
			tr.elem('td', text="<empty>")
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()

# ===========================================================================

class LinearComponent0(TouchNotify):
	def __init__(self, *args, param=None, data=None, scale=None, **kwargs):
		if len(args):
			if isinstance(args[0], LinearComponent0):
				super().__init__(**kwargs)
				self._param = param or args[0]._param
				self._data = data or args[0]._data
				self._scale = scale or args[0]._scale
				self.set_touch_callback(args[0].get_touch_callback())
			else:
				raise TypeError('init LinearComponent0 with LinearComponent0 or keyword arguments')
		super().__init__(**kwargs)
		self._param = param or '1'
		self._data = data or '1'
		self._scale = scale or 1.0

	@property
	def data(self):
		return DataRef(self._data)

	@data.setter
	def data(self, value):
		self._data = value

	@property
	def param(self):
		return ParameterRef(self._param)

	@param.setter
	def param(self, value):
		self._param = value

	@property
	def scale(self):
		return self._scale

	@scale.setter
	def scale(self, value):
		self._scale = value

	def __pos__(self):
		return self

	def __add__(self, other):
		if other == ():
			return self
		elif isinstance(other, (LinearFunction, LinearComponent0)):
			return (LinearFunction(touch_callback=self.get_touch_callback()) + self) + other
		elif other == 0:
			return self
		else:
			raise NotImplementedError()

	def __radd__(self, other):
		if other == 0:
			return self
		return other.__add__(self)

	def __mul__(self, other):
		if isinstance(other, (int, float, )):
			return LinearComponent0(data=self.data, param=self.param, scale=self.scale * other, touch_callback=self.get_touch_callback())
		elif isinstance(other, (DataRef, )):
			return LinearComponent0(data=self.data * other, param=self.param, scale=self.scale, touch_callback=self.get_touch_callback())
		else:
			raise TypeError('unsupported operand type(s) for LinearComponent0: {}'.format(type(other)))

	def __rmul__(self, other):
		if isinstance(other, (int, float, )):
			return LinearComponent0(data=self.data, param=self.param, scale=self.scale * other, touch_callback=self.get_touch_callback())
		elif isinstance(other, (DataRef, )):
			return LinearComponent0(data=self.data * other, param=self.param, scale=self.scale, touch_callback=self.get_touch_callback())
		else:
			raise TypeError('unsupported operand type(s) for LinearComponent0: {}'.format(type(other)))

	def __imul__(self, other):
		if isinstance(other, (int, float, )):
			self.scale = self.scale * other
		elif isinstance(other, (DataRef, )):
			self.data = self.data * other
		else:
			raise TypeError('unsupported operand type(s) for LinearComponent0: {}'.format(type(other)))

	def __iter__(self):
		return iter(LinearFunction(touch_callback=self.get_touch_callback()) + self)

	def __eq__(self, other):
		if isinstance(other, LinearFunction) and len(other) == 1:
			other = other[0]
		if not isinstance(other, LinearComponent0):
			return False
		if self.param != other.param:
			return False
		if self.data != other.data:
			return False
		if self.scale != other.scale:
			return False
		return True

	def __repr__(self):
		try:
			if self.scale == 1.0:
				return f"({self.param!r} * {self.data!r})"
			return f"({self.param!r} * {self.scale} * {self.data!r})"
		except AttributeError:
			return f"<LinearComponent0 {id(self)} with error>"

	def __xml__(self, exponentiate_parameter=False, resolve_parameters=None):
		from xmle import Elem
		x = Elem('div')
		# x << tooltipped_style()

		if resolve_parameters is not None:
			if exponentiate_parameter:
				plabel = resolve_parameters.pvalue(self.param, default_value="This is a Parameter")
				if not isinstance(plabel, str):
					plabel = "exp({0:.3g})={1:.3g}".format(plabel, numpy.exp(plabel))
			else:
				plabel = "{0:.3g}".format(resolve_parameters.pvalue(self.param, default_value="This is a Parameter"))
		else:
			plabel = "This is a Parameter"

		if self.scale == 1.0:
			if exponentiate_parameter:
				x.elem('span', tail="exp(")
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=") * ")
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
			else:
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=" * ")
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
		else:
			if exponentiate_parameter:
				x.elem('span', tail="exp(")
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=f" * {self.scale}) * ")
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
			else:
				a_p = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.param), tail=f" * {self.scale} * ")
				a_p.elem('span', attrib={'class':'tooltiptext'}, text=plabel)
		a_x = x.elem('div', attrib={'class':'tooltipped'}, text=repr(self.data))
		a_x.elem('span', attrib={'class':'tooltiptext'}, text="This is Data")
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()

	def _evaluate(self, p_getter, x_namespace=None, **kwargs):
		return self._scale * p_getter(str(self.param)) * self.data.eval(namespace=x_namespace, **kwargs)

	def __copy__(self):
		return LinearComponent0(
			param=self._param,
			data=self._data,
			scale=self._scale,
		)

	def __deepcopy__(self, memodict):
		return LinearComponent0(
			param=copy.deepcopy(self._param if isinstance(self._param, str) else str(self._param), memodict),
			data=copy.deepcopy(self._data if isinstance(self._data, str) else str(self._data), memodict),
			scale=copy.deepcopy(self._scale, memodict),
		)

# ===========================================================================

class LinearFunction0(TouchNotify, list):
	def __init__(self, *args, touch_callback=None, **kwargs):
		if len(args) == 1 and args[0] is None:
			args = ()
		super().__init__(*args, **kwargs)
		self.set_touch_callback(touch_callback)

	def __add__(self, other):
		if other == ():
			return self
		if isinstance(other, LinearFunction0):
			return type(self)(super().__add__(other), touch_callback=self.get_touch_callback())
		if isinstance(other, ParameterRef):
			other = LinearComponent0(param=other, touch_callback=self.get_touch_callback())
		if isinstance(other, LinearComponent0):
			result = type(self)(self, touch_callback=self.get_touch_callback())
			result.append(other)
			return result
		raise TypeError("cannot add type {} to LinearFunction0".format(type(other)))

	def __iadd__(self, other):
		if isinstance(other, ParameterRef):
			other = LinearComponent0(param=other, touch_callback=self.get_touch_callback())
		if other == ():
			return self
		elif isinstance(other, LinearFunction0):
			super().__iadd__(other)
		elif isinstance(other, LinearComponent0):
			super().append(other)
		else:
			raise TypeError("cannot add type {} to LinearFunction0".format(type(other)))
		return self

	def __radd__(self, other):
		return LinearFunction0(touch_callback=self.get_touch_callback()) + other + self

	def __pos__(self):
		return self

	def __mul__(self, other):
		trial = LinearFunction0(touch_callback=self.get_touch_callback())
		for component in self:
			trial += component * other
		return trial

	def __rmul__(self, other):
		trial = LinearFunction0(touch_callback=self.get_touch_callback())
		for component in self:
			trial += other * component
		return trial

	# def evaluate(self, dataspace, model):
	# 	if len(self) > 0:
	# 		i = self[0]
	# 		y = i.data.eval(**dataspace) * i.param.default_value(0).value(model)
	# 	for i in self[1:]:
	# 		y += i.data.eval(**dataspace) * i.param.default_value(0).value(model)
	# 	return y
	#
	# def evaluator1d(self, factorlabel='', x_ident=None):
	# 	if x_ident is None:
	# 		try:
	# 			x_ident = self._x_ident
	# 		except AttributeError:
	# 			raise TypeError('a x_ident must be given')
	# 	if x_ident is None:
	# 		raise TypeError('a x_ident must be given')
	# 	from .util.plotting import ComputedFactor
	# 	return ComputedFactor(label=factorlabel, func=lambda x, m: self.evaluate({x_ident: x}, m))

	def __contains__(self, val):
		if isinstance(val, ParameterRef):
			for i in self:
				if i.param == val:
					return True
			return False
		if isinstance(val, DataRef):
			for i in self:
				if i.data == val:
					return True
			return False
		raise TypeError("the searched for content must be of type ParameterRef or DataRef")

	def _index_of(self, val):
		if isinstance(val, ParameterRef):
			for n, i in enumerate(self):
				if i.param == val:
					return n
			raise KeyError('ParameterRef not found')
		if isinstance(val, DataRef):
			for n, i in enumerate(self):
				if i.data == val:
					return n
			raise KeyError('DataRef not found')
		raise TypeError("the searched for content must be of type ParameterRef or DataRef")

	def reformat_param(self, container=None, pattern=None, repl=None, **kwargs):
		"""
		Transform all the parameters in the LinearFunction0.

		Parameters
		----------
		container : str
			A format string, into which the previous parameters are formatted.
			Use this to append things to the parameter names.
		pattern : str
		repl : str
			Passed to `re.sub` with each existing parameter as the base string
			to be searched.

		Examples
		--------
		>>> from larch.roles import P,X
		>>> f = P.InVehTime * X.IVTT + P.OutOfVehTime * X.OVTT
		>>> f1 = f.reformat_param('{}_Suffix')
		>>> str(f1)
		'(P.InVehTime_Suffix * X.IVTT) + (P.OutOfVehTime_Suffix * X.OVTT)'
		>>> f2 = f.reformat_param(pattern='(Veh)', repl='Vehicle')
		>>> str(f2)
		'(P.InVehicleTime * X.IVTT) + (P.OutOfVehicleTime * X.OVTT)'

		"""
		import re
		r = LinearFunction0(touch_callback=self.get_touch_callback())
		for i in self:
			if pattern is None:
				param = i.param
			else:
				if repl is None:
					raise TypeError('must give repl with pattern')
				param = re.sub(pattern, repl, i.param, **kwargs)
			if container is None:
				container = '{}'
			r += LinearComponent0(data=i.data, param=container.format(param), touch_callback=self.get_touch_callback())
		try:
			r._x_ident = self._x_ident
		except AttributeError:
			pass
		return r

	def reformat_data(self, container=None, pattern=None, repl=None, **kwargs):
		"""
		Transform all the data in the LinearFunction0.

		Parameters
		----------
		container : str
			A format string, into which the previous data strings are formatted.
			Use this to apply common global transforms to the data.
		pattern : str
		repl : str
			Passed to `re.sub` with each existing data string as the base string
			to be searched.
		"""
		import re
		r = LinearFunction0(touch_callback=self.get_touch_callback())
		for i in self:
			if pattern is None:
				data = i.data
			else:
				if repl is None:
					raise TypeError('must give repl with pattern')
				data = re.sub(pattern, repl, i.data, **kwargs)
			if container is None:
				container = '{}'
			r += LinearComponent0(data=container.format(data), param=i.param, touch_callback=self.get_touch_callback())
		try:
			r._x_ident = self._x_ident
		except AttributeError:
			pass
		return r

	def __code__(self):
		return " + ".join(f"({repr(i)})" for i in self)

	def __eq__(self, other):
		if not isinstance(other, LinearFunction0):
			return False
		if len(self) != len(other):
			return False
		for i, j in zip(self, other):
			if i != j: return False
		return True

	def __getstate__(self):
		state = {}
		state['code'] = self.__code__()
		try:
			state['_x_ident'] = self._x_ident
		except AttributeError:
			pass
		return state

	def __setstate__(self, state):
		self.__init__()
		if state['code']:
			self.__iadd__(eval(state['code']))
		if '_x_ident' in state:
			self._x_ident = state['_x_ident']

	def __repr__(self):
		result = " + ".join(repr(i) for i in self)
		if len(result)<80:
			return result
		else:
			return "  "+result.replace(" + ","\n+ ")

	def __xml__(self, linebreaks=False, lineprefix="", exponentiate_parameters=False, resolve_parameters=None):
		from xmle import Elem
		x = Elem('div', attrib={'class':'LinearFunc'})
		for n,i in enumerate(self):
			ix_ = list(i.__xml__(exponentiate_parameter=exponentiate_parameters, resolve_parameters=resolve_parameters))
			if linebreaks:
				if n>0 or lineprefix:
					ix_.insert(0,Elem('br', tail = lineprefix+" + "))
			else:
				if n < len(self)-1:
					if ix_[-1].tail is None:
						ix_[-1].tail = " + "
					else:
						ix_[-1].tail += " + "
			for ii in ix_:
				x << ii
		return x

	def _repr_html_(self):
		return self.__xml__().tostring()

	def data(self, cls=None):
		if cls is None:
			return [_.data for _ in self]
		else:
			return [cls(_.data) for _ in self]

	def set_touch_callback(self, callback):
		super().set_touch_callback(callback)
		for i in self:
			i.set_touch_callback(callback)

	def _evaluate(self, p_getter, x_namespace=None, **more_x_namespace):
		if hasattr(p_getter,'pvalue') and callable(p_getter.pvalue):
			p_getter = p_getter.pvalue
		return sum(j._evaluate(p_getter, x_namespace=x_namespace, **more_x_namespace) for j in self)

	def copy(self):
		result = LinearFunction0() + self
		return result

	def __deepcopy__(self, memodict):
		result = LinearFunction0()
		for i in self:
			result.append(copy.deepcopy(i, memodict))
		return result

	def _linear_plot_2d_data(self, p_getter, x_name, x_min, x_max, n_points=100, **other_namespace):
		import numpy
		if hasattr(self, 'plotting_namespace') and len(other_namespace)==0:
			other_namespace = self.plotting_namespace
		x = numpy.linspace(x_min, x_max, n_points)
		y = self._evaluate(p_getter, {x_name:x}, **other_namespace)
		return x,y

	def linear_plot_2d(self, p_getter, x_name, x_min, x_max, n_points=100, *, xlabel=None, svg=True, header=None, **other_namespace):

		# Delayed evaluation mode...
		if p_getter is None:
			return lambda x: self.linear_plot_2d(x, x_name, x_min, x_max, n_points=n_points, xlabel=xlabel, svg=svg, header=header, **other_namespace)

		# Active evaluation mode...
		from matplotlib import pyplot as plt
		plt.clf()
		x,y = self._linear_plot_2d_data(p_getter, x_name, x_min, x_max, n_points, **other_namespace)
		if hasattr(self, 'plotting_label'):
			plt.plot(x, y, label=self.plotting_label)
		else:
			plt.plot(x, y)
		if xlabel is None:
			plt.xlabel(x_name)
		else:
			plt.xlabel(xlabel)
		plt.tight_layout(pad=0.5)
		from .util.plotting import plot_as_svg_xhtml
		if svg is True:
			svg = {}
		if svg or svg == {}:
			if header is not None:
				svg['header'] = header
			return plot_as_svg_xhtml(plt, **svg)
		else:
			plt.show()

	def _inplot_linear_plot_2d(self, plt, p_getter, x_name, x_min, x_max, n_points=100, *, xlabel=None, svg=True, header=None, **other_namespace):

		# Delayed evaluation mode...
		if p_getter is None:
			return lambda x: self._inplot_linear_plot_2d(plt, x, x_name, x_min, x_max, n_points=n_points, xlabel=xlabel, svg=svg, header=header, **other_namespace)

		# Active evaluation mode...
		x,y = self._linear_plot_2d_data(p_getter, x_name, x_min, x_max, n_points, **other_namespace)
		if hasattr(self, 'plotting_label'):
			plt.plot(x, y, label=self.plotting_label)
		else:
			plt.plot(x,y)

	def total_ordering_increasing(self):
		from toolz.itertoolz import sliding_window
		snaps = []
		for windows_size in range(2, len(self)-1):
			for sub_p in sliding_window(windows_size, self):
				snaps.append(sub_p[0].param.lessthan(sub_p[-1].param))

		return snaps

# ====================================================================================================
def multiple_linear_plot_2d(linear_funcs, p_getter, x_name, x_min, x_max, n_points=100, *, xlabel=None, svg=True, header=None, other_namespace=None):

	# Delayed evaluation mode...
	if p_getter is None:
		return lambda x: multiple_linear_plot_2d(linear_funcs, x, x_name, x_min, x_max, n_points=n_points, xlabel=xlabel,
													 svg=svg, header=header, other_namespace=other_namespace)

	# Active evaluation mode...
	from matplotlib import pyplot as plt
	plt.clf()

	if other_namespace is None:
		for lf in linear_funcs:
			lf._inplot_linear_plot_2d(plt, p_getter, x_name, x_min, x_max, n_points, xlabel=xlabel, svg=svg, header=header)
	else:
		for lf, oname in zip(linear_funcs, other_namespace):
			lf._inplot_linear_plot_2d(plt, p_getter, x_name, x_min, x_max, n_points, xlabel=xlabel, svg=svg, header=header, **oname)

	if xlabel is not None:
		plt.xlabel(xlabel)

	plt.legend()

	plt.tight_layout(pad=0.5)
	from .util.plotting import plot_as_svg_xhtml
	if svg is True:
		svg = {}
	if svg or svg == {}:
		if header is not None:
			svg['header'] = header
		return plot_as_svg_xhtml(plt, **svg)
	else:
		plt.show()


# ====================================================================================================

from itertools import chain
_RaiseKeyError = object() # singleton for no-default behavior

class DictOfLinearFunction0(TouchNotify, dict):
	@staticmethod # because this doesn't make sense as a global function.
	def _process_args(mapping=(), **kwargs):
		if hasattr(mapping, 'items'):
			mapping = getattr(mapping, 'items')()
		return ((k, v) for k, v in chain(mapping, getattr(kwargs, 'items')()))
	def __init__(self, mapping=(), touch_callback=None, alts_validator=None, **kwargs):
		if mapping is None:
			mapping = ()
		super().__init__(self._process_args(mapping, **kwargs))
		self.set_touch_callback(touch_callback if callable(touch_callback) else lambda: None)
		self._alts_validator = alts_validator
	def set_alts_validator(self, av):
		self._alts_validator = av
	def __getitem__(self, k):
		try:
			return super().__getitem__(k)
		except KeyError:
			if self._alts_validator is None or self._alts_validator(k):
				super().__setitem__(k, LinearFunction0(touch_callback=self.get_touch_callback()))
				return super().__getitem__(k)
			else:
				raise
	def __setitem__(self, k, v):
		if isinstance(v, ParameterRef):
			v = v * DataRef('1')
		if isinstance(v, int) and v==0:
			v = LinearFunction0(touch_callback=self.get_touch_callback())
		elif isinstance(v, LinearComponent0):
			v = LinearFunction0(touch_callback=self.get_touch_callback()) + v
		elif isinstance(v, LinearFunction0):
			v.set_touch_callback(self.get_touch_callback())
		elif isinstance(v, list):
			v_ = LinearFunction0(touch_callback=self.get_touch_callback())
			for vv in v:
				v_ = v_ + vv
			v = v_
		else:
			raise TypeError("only accepts LinearFunction0 values")
		if "with error" in repr(v):
			raise ValueError("found error here")
		return super().__setitem__(k, v)
	def setdefault(self, k, default=None):
		if default is None:
			default = LinearFunction0(touch_callback=self.get_touch_callback())
		if isinstance(default, LinearComponent0):
			default = LinearFunction0(touch_callback=self.get_touch_callback()) + default
		elif isinstance(default, LinearFunction0):
			default.set_touch_callback(self.get_touch_callback())
		else:
			raise TypeError("only accepts LinearFunction0 values")
		return super().setdefault(k, default)
	def copy(self): # don't delegate w/ super - dict.copy() -> dict :(
		return type(self)(self)
	def copy_without_touch_callback(self, deep=True):
		result = type(self)()
		if deep:
			for k in self.keys():
				result[k] = list(self[k])
		else:
			for k in self.keys():
				result[k] = self[k]
		return result
	def __repr__(self):
		return '{0}({1})'.format(type(self).__name__, super().__repr__())
	def set_touch_callback(self, callback):
		super().set_touch_callback(callback)
		for i in self.values():
			i.set_touch_callback(callback)
	def __xml__(self):
		from pprint import pformat
		from xmle import Elem
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;")
		t.elem('caption', text=f"<larch.{self.__class__.__name__}>", style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;")
		if len(self):
			tr = t.elem('tr')
			tr.elem('th', text="alt")
			tr.elem('th', text='formula')
			for k,v in self.items():
				tr = t.elem('tr')
				tr.elem('td', text=str(k))
				try:
					v_ = v.__xml__()
				except AttributeError:
					tr.elem('td', text=str(v))
				else:
					tr.elem('td') << v_
		else:
			tr = t.elem('tr')
			tr.elem('td', text="<empty>")
		return x
	def _repr_html_(self):
		return self.__xml__().tostring()
#
#
#

class DictOfStrings(TouchNotify, dict):
	@staticmethod # because this doesn't make sense as a global function.
	def _process_args(mapping=(), **kwargs):
		if hasattr(mapping, 'items'):
			mapping = getattr(mapping, 'items')()
		return ((k, v) for k, v in chain(mapping, getattr(kwargs, 'items')()))
	def __init__(self, mapping=(), touch_callback=None, **kwargs):
		if mapping is None:
			mapping = ()
		super().__init__(self._process_args(mapping, **kwargs))
		self.set_touch_callback(touch_callback if callable(touch_callback) else lambda: None)
	def __contains__(self, k):
		return super().__contains__(str(k))
	def __getitem__(self, k):
		return super().__getitem__(str(k))
	def __setitem__(self, k, v):
		return super().__setitem__(str(k), str(v))
	def setdefault(self, k, default=""):
		return super().setdefault(str(k), str(default))
	def copy(self): # don't delegate w/ super - dict.copy() -> dict :(
		return type(self)(self)
	def __repr__(self):
		return '{0}({1})'.format(type(self).__name__, super().__repr__())
	def set_touch_callback(self, callback):
		super().set_touch_callback(callback)
	def __xml__(self):
		from pprint import pformat
		from xmle import Elem
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;")
		t.elem('caption', text=f"<larch.{self.__class__.__name__}>", style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;")
		if len(self):
			tr = t.elem('tr')
			tr.elem('th', text="key")
			tr.elem('th', text='value')
			for k,v in self.items():
				tr = t.elem('tr')
				tr.elem('td', text=str(k))
				tr.elem('td', text=str(v))
		else:
			tr = t.elem('tr')
			tr.elem('td', text="<empty>")
		return x
	def _repr_html_(self):
		return self.__xml__().tostring()

class DictOfStringKeys(TouchNotify, dict):
	@staticmethod # because this doesn't make sense as a global function.
	def _process_args(mapping=(), **kwargs):
		if hasattr(mapping, 'items'):
			mapping = getattr(mapping, 'items')()
		return ((k, v) for k, v in chain(mapping, getattr(kwargs, 'items')()))
	def __init__(self, mapping=(), touch_callback=None, **kwargs):
		if mapping is None:
			mapping = ()
		super().__init__(self._process_args(mapping, **kwargs))
		self.set_touch_callback(touch_callback if callable(touch_callback) else lambda: None)
	def __contains__(self, k):
		return super().__contains__(str(k))
	def __getitem__(self, k):
		return super().__getitem__(str(k))
	def __setitem__(self, k, v):
		return super().__setitem__(str(k), v)
	def setdefault(self, k, default=""):
		return super().setdefault(str(k), default)
	def copy(self): # don't delegate w/ super - dict.copy() -> dict :(
		return type(self)(self)
	def __repr__(self):
		return '{0}({1})'.format(type(self).__name__, super().__repr__())
	def set_touch_callback(self, callback):
		super().set_touch_callback(callback)
	def __xml__(self):
		from pprint import pformat
		from xmle import Elem
		x = Elem('div')
		t = x.elem('table', style="margin-top:1px;")
		t.elem('caption', text=f"<larch.{self.__class__.__name__}>", style="caption-side:top;text-align:left;font-family:Roboto;font-weight:700;font-style:normal;font-size:100%;padding:0px;")
		if len(self):
			tr = t.elem('tr')
			tr.elem('th', text="key")
			tr.elem('th', text='value')
			for k,v in self.items():
				tr = t.elem('tr')
				tr.elem('td', text=str(k))
				tr.elem('td', text=str(v))
		else:
			tr = t.elem('tr')
			tr.elem('td', text="<empty>")
		return x
	def _repr_html_(self):
		return self.__xml__().tostring()



from .model.linear import ParameterRef_C, DataRef_C, LinearFunction_C, \
	LinearComponent_C, DictOfLinearFunction_C, Ref_Gen

LinearFunction = LinearFunction_C
LinearComponent = LinearComponent_C
DictOfLinearFunction = DictOfLinearFunction_C

P = Ref_Gen(ParameterRef_C)
X = Ref_Gen(DataRef_C)

def PX(z):
	return P(z) * X(z)

globals()[_ParameterRef_repr_txt] = P
globals()[_DataRef_repr_txt] = X
globals()[_ParameterRef_repr_txt + _DataRef_repr_txt] = PX

__all__ = [_ParameterRef_repr_txt, _DataRef_repr_txt, _ParameterRef_repr_txt + _DataRef_repr_txt]




from .. import __version__
import sys, os

class Info:

    def __init__(self, appname='Larch', extra=True, version=None, path=None, minimal=False):
        self.appname = appname
        self.extra = extra
        self.version = version or __version__
        self.minimal = minimal
        from .. import __path__
        self.path = path or __path__[0]


    def __repr__(self):
        r = (f"┌── {self.appname.upper()} {self.version} " + "─" * (57 - len(self.version)))
        v = '\n│'.join(sys.version.split('\n'))
        r += (f"\n│Python {v}")
        r += (f"\n│EXE ─ {sys.executable}")
        r += (f"\n│CWD ─ {os.getcwd()}" )
        for p in sys.path[:1]:
            r += (f"\n│PTH ┬ {p}")
        for p in sys.path[1:-1]:
            r += (f"\n│    ├ {p}")
        for p in sys.path[-1:]:
            r += (f"\n│    └ {p}")
        r += ("\n└───────────────────────────────────────────────────────────────────────────")
        return r

    def _repr_html_(self):
        from ..util.styles import _default_css_jupyter, _tooltipped_style_css
        style_prefix = "<style>{}\n\n{}</style>\n".format(_default_css_jupyter, _tooltipped_style_css)
        from xmle import Elem
        xsign = Elem("div", {'class': 'larch_head_tag'})
        from .images import favicon
        p = xsign.elem('p', {'style': 'float:left;margin-top:6px'})
        p.elem('img', {
            'width': "32",
            'height': "32",
            'src': "data:image/png;base64,{}".format(favicon),
            'style': 'float:left;position:relative;top:-3px;padding-right:0.2em;'
        }, tail=f" {self.appname} ")
        p.elem('span', {'class': 'larch_head_tag_ver'}, text=f" {self.version} ")
        if not self.minimal:
            p.elem('span', {'class': 'larch_head_tag_pth'}, text=f" {self.path} ")
            from .images import camsyslogo_element
            xsign << camsyslogo_element
            if 'larch' in sys.modules:
                from .images import georgiatechlogo_element
                xsign << georgiatechlogo_element

            if self.extra:
                v = '\n│'.join(sys.version.split('\n'))
                xsign.elem('br')
                xinfo = xsign.elem('div', {'class': 'larch_head_tag_more', 'style':'margin-top:10px; padding:7px'}, text=f'Python {v}')
                xinfo.elem('br', tail=f"EXE - {sys.executable}")
                xinfo.elem('br', tail=f"CWD - {os.getcwd()}")
                xinfo.elem('br', tail=f"PATH - ")
                ul = xinfo.elem('ul', {'style': 'margin-top:0; margin-bottom:0;'})
                for p in sys.path:
                    ul.elem('li', text=p)

        return style_prefix+xsign.tostring()


    def __call__(self, *args, **kwargs):
        """
        Calling an Info object is a no-op.

        Returns
        -------
        self
        """
        return self




def ipython_status(magic_matplotlib=False):
    message_set = set()
    try:
        # This will work in iPython, and fail otherwise
        cfg = get_ipython().config
    except:
        message_set.add('Not IPython')
    else:
        import IPython
        import IPython.core.error
        message_set.add('IPython')

        if magic_matplotlib:
            try:
                get_ipython().magic("matplotlib inline")
            except (IPython.core.error.UsageError, KeyError):
                message_set.add('IPython inline plotting not available')

        # Caution: cfg is an IPython.config.loader.Config
        if cfg['IPKernelApp']:
            message_set.add('IPython QtConsole')
            try:
                if cfg['IPKernelApp']['pylab'] == 'inline':
                    message_set.add('pylab inline')
                else:
                    message_set.add('pylab loaded but not inline')
            except:
                message_set.add('pylab not loaded')
        elif cfg['TerminalIPythonApp']:
            try:
                if cfg['TerminalIPythonApp']['pylab'] == 'inline':
                    message_set.add('pylab inline')
                else:
                    message_set.add('pylab loaded but not inline')
            except:
                message_set.add('pylab not loaded')
    return message_set





from .generic import *
from ...general import _sqz_same, _sqz_same_trailing_neg_ok
import warnings

class H5PodCO(H5Pod):
	"""
	A HDF5 group node containing :ref:`idco` format data.

	The CO pod is the most basic H5Pod. It includes a selection of data columns, where
	each column is a vector of values attributable to casewise observations.  The ordering
	of the data is important, and all data columns in an H5PodCO must have the same length
	and be ordered the same way, so that any group of columns can be simply concantenated to
	create a 2-d :ref:`idco` array of data.
	"""
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

	@property
	def podtype(self):
		return 'idco'


	@classmethod
	def from_dataframe(cls, df, h5filename=None, h5mode='a', h5groupnode='/',
					   inmemory=False, temp=False, complevel=1, complib='zlib', ident=None):
		self = cls(h5filename, h5mode, h5groupnode, inmemory=inmemory, temp=temp, complevel=complevel, complib=complib, ident=ident)
		self.merge_external_data(df, self_on=None, other_on=None)
		return self

	@classmethod
	def from_csv(cls, *args, h5filename=None, h5mode='a', h5groupnode='/',
				 inmemory=False, temp=False, complevel=1, complib='zlib', ident=None,
				 rename_columns=None, force_natural_names=True,
				 **kwargs):
		self = cls(h5filename, h5mode, h5groupnode, inmemory=inmemory, temp=temp, complevel=complevel, complib=complib, ident=ident)
		df = pandas.read_csv(*args, **kwargs)
		if rename_columns:
			df.columns = [(rename_columns[c] if c in rename_columns else c) for c in df.columns]
		df.columns = [c.replace("/",'_') for c in df.columns]
		if force_natural_names:
			from ....util.naming import make_valid_identifier
			df.columns = [make_valid_identifier(c, True) for c in df.columns]
		self.merge_external_data(df, self_on=None, other_on=None)
		return self

	@classmethod
	def from_excel(cls, *args, h5filename=None, h5mode='a', h5groupnode='/',
				   inmemory=False, temp=False, complevel=1, complib='zlib', ident=None,
				   rename_columns=None, force_natural_names=True,
				   **kwargs):
		self = cls(h5filename, h5mode, h5groupnode, inmemory=inmemory, temp=temp, complevel=complevel, complib=complib, ident=ident)
		df = pandas.read_excel(*args, **kwargs)
		if rename_columns:
			df.columns = [(rename_columns[c] if c in rename_columns else c) for c in df.columns]
		df.columns = [c.replace("/",'_') for c in df.columns]
		if force_natural_names:
			from ....util.naming import make_valid_identifier
			df.columns = [make_valid_identifier(c, True) for c in df.columns]
		self.merge_external_data(df, self_on=None, other_on=None)
		return self


	def copy_external_data(self, other, names=None, original_source=None,
						log=lambda *x: None, **kwargs):
		"""
		Copy data into this H5PodCO.

		Every case in the current (receiving) H5PodCO should match exactly one case in the
		imported data, in the same order, so that no merge is undertaken; the data is simplied copied.

		Parameters
		----------
		other : H5PodCO
			The other data table to be merged.  Can be another H5PodCO, or a DataFrame, or the
			path to a file of type {csv, xlsx, dbf}.
		names : iterable or dict, optional
			If given as a list, only these names will be merged from the other data.
			If given as a dict, the keys are the names that will be merged and the
			values are the new names in this DT.
		original_source : str, optional
			Give the original source of this data.  If not given and the filename can be
			inferred from other, that name will be used.
		"""
		if names is not None and not isinstance(names, dict):
			if isinstance(names, str):
				names = {names: names}
			else:
				names = {n: n for n in names}
		if not isinstance(other, H5PodCO):
			raise TypeError("currently can merge copy H5PodCO")
		if original_source is None:
			original_source = other._h5f.filename
		anything_imported = False
		for col in other.names():
			if names is None and col not in self._groupnode:
				log('importing "{}" into {}'.format(col, self.filename))
				self.add_array(col, arr=other._groupnode._v_children[col][:])
				anything_imported = True
				if original_source is not None:
					self._groupnode._v_children[col]._v_attrs.ORIGINAL_SOURCE = original_source
				if col in other._groupnode and 'DICTIONARY' in other._groupnode._v_children[col]._v_attrs:
					self._groupnode._v_children[col]._v_attrs.DICTIONARY = other._groupnode._v_children[col]._v_attrs.DICTIONARY
			elif names is not None and col in names and names[col] not in self._groupnode:
				log('importing "{}" as "{}" into {}'.format(col, names[col], self.filename))
				self.add_array(names[col], arr=other._groupnode._v_children[col][:])
				anything_imported = True
				if original_source is not None:
					self._groupnode._v_children[names[col]]._v_attrs.ORIGINAL_SOURCE = original_source
				if col in other._groupnode and 'DICTIONARY' in other._groupnode._v_children[col]._v_attrs:
					self._groupnode._v_children[names[col]]._v_attrs.DICTIONARY = other._groupnode._v_children[col]._v_attrs.DICTIONARY
		if not anything_imported:
			if names is not None:
				names_warn = '\n  from names:\n    ' + '\n    '.join(names.keys())
			else:
				names_warn = ''
			warnings.warn('nothing was imported into {}{}'.format(self.filename, names_warn))


	def merge_external_data(
			self, other,
			self_on, other_on=None, other_on_index=False,
			dupe_suffix="_copy", original_source=None, names=None, overwrite=False,
			other_dedupe=False, log=lambda *x: None, **kwargs
	):
		"""
		Merge data into this H5PodCO.

		Every case in the current (receiving) H5PodCO should match one or zero cases in the
		imported data.

		Parameters
		----------
		other : H5PodCO or pandas.DataFrame or str
			The other data table to be merged.  Can be another DT, or a DataFrame, or the
			path to a file of type {csv, xlsx, dbf}.
		self_on : label or list, or array-like
			Field names to join on in this H5PodCO. Can be a vector or list of vectors
			of correct length to use a particular vector as the join key instead of columns
		other_on : label or list, or array-like
			Field names to join on in other DataFrame, or vector/list of vectors per self_on
		dupe_suffix : str
			A suffix to add to variables that duplicate variables names already in this H5PodCO
		original_source : str, optional
			Give the original source of this data.  If not given and the filename can be
			inferred from other, that name will be used.
		names : iterable or dict, optional
			If given as a list, only these names will be merged from the other data.
			If given as a dict, the keys are the names that will be merged and the
			values are the new names in this DT.
		"""
		if other_on is None:
			if not other_on_index:
				other_on = self_on
			else:
				other_on = []

		if names is not None and not isinstance(names, dict):
			if isinstance(names, str):
				names = {names: names}
			else:
				names = {n: n for n in names}

		def _warn_nothing_done_():
			if names is not None:
				names_warn = '\n  from names:\n    ' + '\n    '.join(names.keys())
			else:
				names_warn = ''
			warnings.warn('nothing was imported into {}{}'.format(self.filename, names_warn))

		# if isinstance(other, pandas.DataFrame):
		# 	return self.merge_into_idco_from_dataframe(other, self_on, other_on, dupe_suffix=dupe_suffix,
		# 											   original_source=original_source, names=names)
		# if isinstance(other, str) and os.path.exists(other):
		# 	return self.merge_into_idco_from_csv(other, self_on, other_on, dupe_suffix=dupe_suffix,
		# 										 original_source=original_source, names=names, **kwargs)

		if isinstance(self_on, str):
			self_on = (self_on,)
		if isinstance(other_on, str):
			other_on = (other_on,)

		if self_on is None:
			baseframe = None
		elif isinstance(self_on, str):
			baseframe = self.dataframe[self_on]
		else:
			baseframe = self.dataframe[tuple(self_on)]
		anything_imported = False

		# get the other data (to be merged) as a pandas DataFrame
		if isinstance(other, H5PodCO):
			# we have a H5PodCO
			og = other._groupnode
			if other_on_index:
				raise TypeError('other_on_index cannot be True for other as H5PodCO')
			if original_source is None:
				original_source = other._h5f.filename
			if names is None:
				# we are potentially importing everything, so load everything into memory
				other_df = other.dataframe[other.names()]
			else:
				# we are importing only some things, so load only what will be needed
				use_names = set(names.keys()) | (set(other_on))
				for col in other.names():
					if col in other_on:
						continue
					if col not in names or (names[col] in self._groupnode and not overwrite):
						use_names.discard(col)
				if use_names == set(other_on):
					# there is nothing left to load but the matching keys,
					# so nothing will be loaded.  Abort now and save a bunch of work.
					_warn_nothing_done_()
					return
				other_df = other.dataframe[use_names]
		elif isinstance(other, pandas.DataFrame):
			# it is already a dataframe, so convenient!
			other_df = other
			og = None
		else:
			raise TypeError("bad type for other data")

		if other_dedupe:
			try:
				other_df.drop_duplicates(
					subset=None if other_on_index else other_on,
					keep='first',
					inplace=True
				)
			except TypeError as err:
				raise ValueError("Drop Duplicates Error") from err

		if baseframe is not None:
			new_df = pandas.merge(
				baseframe,
				other_df,
				left_on=self_on,
				right_on=None if other_on_index else other_on,
				right_index=other_on_index,
				how='left',
				suffixes=('', dupe_suffix),
			)
		else:
			new_df = other_df

		log(new_df)
		for col in new_df.columns:
			if names is None and col not in self._groupnode:
				log('importing "{}" into {}'.format(col, self.filename))
				try:
					self.add_array(col, arr=new_df[col].values)
				except UnicodeEncodeError:
					# The datatype has some problems converting to H5, warn and skip it.
					warnings.warn(f"unable to import {col}, the datatype is weird")
					continue
				anything_imported = True
				if original_source is not None:
					self._groupnode._v_children[col]._v_attrs.ORIGINAL_SOURCE = original_source
				if og is not None and col in og and 'DICTIONARY' in og._v_children[col]._v_attrs:
					self._groupnode._v_children[col]._v_attrs.DICTIONARY = og._v_children[col]._v_attrs.DICTIONARY
			elif names is not None and col in names and (names[col] not in self._groupnode or overwrite):
				log('importing "{}" as "{}" into {}'.format(col, names[col], self.filename))
				try:
					self.add_array(names[col], arr=new_df[col].values)
				except UnicodeEncodeError:
					# The datatype has some problems converting to H5, warn and skip it.
					warnings.warn(f"unable to import {col}, the datatype is weird")
					continue
				anything_imported = True
				if original_source is not None:
					self._groupnode._v_children[names[col]]._v_attrs.ORIGINAL_SOURCE = original_source
				if og is not None and col in og and 'DICTIONARY' in og._v_children[col]._v_attrs:
					self._groupnode._v_children[names[col]]._v_attrs.DICTIONARY = og._v_children[col]._v_attrs.DICTIONARY
		if not anything_imported:
			_warn_nothing_done_()

	def pluck_from_omx(self, other_omx, rowindexes, colindexes, names=None, overwrite=False):
		"""
		Pluck values from an OMX file into new :ref:`idco` variables.

		This method takes O and D index numbers and plucks the individual matrix values at
		those coordinates. New idco variables will be created in the DT file that contain
		the plucked values, so that the new variables represent actual arrays and not links to
		the original matrix.  The OMX filename is marked as the original source of the data.

		Parameters
		----------
		other_omx : OMX or str
			Either an OMX or a filename to an OMX file.
		rowindexes, colindexes : array
			Zero-based index array for the row (origins) and columns (destinations) that will
			be plucked into the new variable.
		names : str or list or dict
			If a str, only that single named matrix table in the OMX will be plucked.
			If a list, all of the named matrix tables in the OMX will be plucked.
			If a dict, the keys give the matrix tables to pluck from and the values
			give the new variable names to create in this DT.


		See Also
		--------
		GroupNode.add_external_omx

		"""
		from ....omx import OMX
		if isinstance(other_omx, str):
			other_omx = OMX(other_omx)
		if names is None:
			names = {n: n for n in other_omx.data._v_children}

		if names is not None and not isinstance(names, dict):
			if isinstance(names, str):
				names = {names: names}
			else:
				names = {n: n for n in names}

		if isinstance(rowindexes, str):
			rowarr = self._groupnode._v_children[rowindexes][:]
		elif isinstance(rowindexes, numpy.ndarray):
			rowarr = rowindexes
		elif isinstance(rowindexes, int) and isinstance(colindexes, (str, numpy.ndarray)):
			rowarr = None  # fix in a moment
		else:
			raise TypeError('rowindexes must be str, int, or ndarray')

		if isinstance(colindexes, str):
			colarr = self._groupnode._v_children[colindexes][:]
		elif isinstance(colindexes, numpy.ndarray):
			colarr = colindexes
		elif isinstance(colindexes, int) and rowarr is not None:
			colarr = numpy.full_like(rowarr, colindexes, dtype=int)
		else:
			raise TypeError('colindexes must be str, int, or ndarray')

		if rowarr is None:
			#rowarr = numpy.full_like(clarr, rowindexes, dtype=int)
			raise NotImplementedError

		for matrix_page in names:
			self.add_array(names[matrix_page],
									 other_omx[matrix_page][rowarr, colarr],
									 original_source=other_omx.filename,
									 overwrite=overwrite)

	def as_idca(self):
		ret = H5PodCOasCA(self, ident=self.ident+"_as_idca")
		return ret

	def fatten(self, fat_filename=None, fat_groupnode=None, fat_dim=None, skips=(), keeps=(), overwrite=False):
		"""Generate a new H5PodCA with data broadcast across a new dimension."""

		if fat_filename is None:
			fat_filename = "{}_fat{}".format(*os.path.splitext(self.filename))

		if fat_filename == self.filename:
			raise ValueError("cannot overwrite same file (maybe one day)")

		if fat_groupnode is None:
			fat_groupnode = self._groupnode._v_pathname

		if fat_dim is None:
			raise TypeError('must give fat_dim')

		from .idca import H5PodCA
		import re
		result = H5PodCA(fat_filename, 'a' if not overwrite else 'w', fat_groupnode)
		for n in self.names():
			keep_n = True
			for s in skips:
				if re.match(s,n):
					keep_n = False
					break
			if len(keeps):
				k_n = [re.match(k, n) for k in keeps]
				if not numpy.any(k_n):
					keep_n = False
			if not keep_n:
				continue
			arr = self.__getattr__(n)[:]
			fat = numpy.empty(arr.shape + (fat_dim,), dtype=arr.dtype)
			fat[..., :] = arr[..., None]
			try:
				di = self.__getattr__(n)._v_attrs.DICTIONARY
			except AttributeError:
				di = None
			try:
				ti = self.__getattr__(n)._v_attrs.TITLE
			except AttributeError:
				ti = None
			result.add_array(n, fat, title=ti, dictionary=di)
		return result

	def reshape_to(self, shape, filename=None, groupnode=None, skips=(), keeps=(), overwrite=False, **kwargs):
		"""Generate a new H5PodCA with data reshaped to fill a new dimension."""

		assert isinstance(shape, (list,tuple))

		if filename is None:
			filename = "{}_ca{}".format(*os.path.splitext(self.filename))

		if filename == self.filename:
			raise ValueError("cannot overwrite same file (maybe one day)")

		if groupnode is None:
			groupnode = self._groupnode._v_pathname

		from .idca import H5PodCA
		import re
		result = H5PodCA(filename, 'a' if not overwrite else 'w', groupnode, **kwargs)
		for n in self.names():
			keep_n = True
			for s in skips:
				if re.match(s,n):
					keep_n = False
					break
			if len(keeps):
				k_n = [re.match(k, n) for k in keeps]
				if not numpy.any(k_n):
					keep_n = False
			if not keep_n:
				continue
			arr = self.__getattr__(n)[:].reshape(shape)
			try:
				di = self.__getattr__(n)._v_attrs.DICTIONARY
			except AttributeError:
				di = None
			try:
				ti = self.__getattr__(n)._v_attrs.TITLE
			except AttributeError:
				ti = None
			result.add_array(n, arr, title=ti, dictionary=di)
		return result

class H5PodCOasCA(H5PodCO):

	@property
	def podtype(self):
		return 'idca'

	def load_data_item(self, name, result, selector=None):
		"""Load a slice of the pod arrays into an array in memory"""
		# convert a single name string to a one item list
		_sqz_same_trailing_neg_ok(result.shape, [selector_len_for(selector, self.shape[0]), *self.shape[1:]])
		result[:,:] = self._evaluate_single_item(name, selector)[:,None]
		return result

	@property
	def shape(self):
		"""The shape of the pod.

		"""
		return super().shape + (self.trailing_dim,)
		# if 'SHAPE' in self._groupnode._v_attrs:
		# 	return tuple(self._groupnode._v_attrs['SHAPE'][:]) + (self.trailing_dim,)
		# if len(self.names()):
		# 	for v in self._groupnode._v_children.values():
		# 		try:
		# 			found_shape = v.shape
		# 		except:
		# 			pass
		# 		else:
		# 			try:
		# 				self.shape = found_shape
		# 			except:
		# 				pass
		# 			return tuple(found_shape) + (self.trailing_dim,)
		# raise NoKnownShape()

	@property
	def trailing_dim(self):
		try:
			return self._trailing_dim
		except AttributeError:
			return -1

	@trailing_dim.setter
	def trailing_dim(self, value):
		self._trailing_dim = int(value)




import warnings
import keyword
import re


class NotAPythonIdentifier(Warning):
	pass


def make_valid_identifier(x, suppress_warnings=False):
	x = str(x)
	if keyword.iskeyword(x):
		y = "_" + x
		if not suppress_warnings:
			warnings.warn("name {0} is a python keyword, converting to {1}".format(x, y), stacklevel=2)
	else:
		y = x
	replacer = re.compile('(\W+)')
	y = replacer.sub("_", y.strip())
	if not y.isidentifier():
		y = "_" + y
	if y != x:
		if not suppress_warnings:
			warnings.warn("name {0} is not a valid python identifier, converting to {1}".format(x, y), stacklevel=2,
			              category=NotAPythonIdentifier)
	return y



def valid_identifier_or_parenthized_string(x, leading_dot=True):
	x = str(x)
	if x.isidentifier():
		if leading_dot:
			return "."+x
		return x
	else:
		if "'" in x and '"' not in x:
			return '("{}")'.format(x)
		if "'" not in x and '"' in x:
			return "('{}')".format(x)
		raise NotImplementedError("cannot handle strings with both quote types")



def parenthize(x, signs_qualify=False):
	"""Wrap a string in parenthesis if needed for unambiguous clarity.

	Parameters
	----------
	x : str
		The string to wrap
	signs_qualify : bool
		If True, a leading + or - on a number triggers the parenthesis (defaults False)

	"""
	x = str(x).strip()
	replacer = re.compile('(\W+)')
	if replacer.search(x):
		if signs_qualify:
			numeric = re.compile('^(([1-9][0-9]*\.?[0-9]*)|(\.[0-9]+))([Ee][+-]?[0-9]+)?\Z')
		else:
			numeric = re.compile('^[+-]?(([1-9][0-9]*\.?[0-9]*)|(\.[0-9]+))([Ee][+-]?[0-9]+)?\Z')
		if numeric.search(x):
			return x
		return "({})".format(x)
	return x



import time
from datetime import datetime

def timesize_single(t):
	if t<60:
		return f"{t:.2f}s"
	elif t<3600:
		return f"{t/60:.2f}m"
	elif t<86400:
		return f"{t/3600:.2f}h"
	else:
		return f"{t/86400:.2f}d"


def timesize_stack(t):
	if t<60:
		return f"{t:.2f}s"
	elif t<3600:
		return f"{t//60:.0f}m {timesize_stack(t%60)}"
	elif t<86400:
		return f"{t//3600:.0f}h {timesize_stack(t%3600)}"
	else:
		return f"{t//86400:.0f}d {timesize_stack(t%86400)}"



class Seconds(float):

	def __repr__(self):
		t = self
		if t < 60:
			return f"{t:.2f}s"
		elif t < 3600:
			return f"{t//60:.0f}m {timesize_stack(t%60)}"
		elif t < 86400:
			return f"{t//3600:.0f}h {timesize_stack(t%3600)}"
		else:
			return f"{t//86400:.0f}d {timesize_stack(t%86400)}"


class Timer:

	def __init__(self):
		self.restart()

	def restart(self):
		self._starttime = datetime.now()
		self._stoptime = None
		return self._starttime

	def start(self):
		return self._starttime

	def stop(self):
		self._stoptime = datetime.now()
		return self._stoptime

	def elapsed(self):
		if self._stoptime:
			return (self._stoptime - self._starttime)
		else:
			return (datetime.now() - self._starttime)

	def __enter__(self):
		self.restart()
		return self

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.stop()




import numpy as np
import pandas as pd
import pandas.core.algorithms as algos

def remove_unused_level(multiindex, level=0):
	"""
	Remove unused labels from a level.

	Create a new MultiIndex from the current that removes
	unused levels from a given level[s], meaning that they are
	not expressed in that level's labels.
	The resulting MultiIndex will have the same outward
	appearance, meaning the same .values and ordering. It will also
	be .equals() to the original.

	Parameters
	----------
	multiindex : pandas.MultiIndex
	level : int or List[int]

	Returns
	-------
	pandas.MultiIndex

	"""

	new_levels = []
	new_codes = []
	if isinstance(level, int):
		level = (level, )

	changed = False
	for n, (lev, level_codes) in enumerate(zip(multiindex.levels, multiindex.codes)):

		if n in level:
			# Since few levels are typically unused, bincount() is more
			# efficient than unique() - however it only accepts positive values
			# (and drops order):
			uniques = np.where(np.bincount(level_codes + 1) > 0)[0] - 1
			has_na = int(len(uniques) and (uniques[0] == -1))

			if len(uniques) != len(lev) + has_na:
				# We have unused levels
				changed = True

				# Recalculate uniques, now preserving order.
				# Can easily be cythonized by exploiting the already existing
				# "uniques" and stop parsing "level_codes" when all items
				# are found:
				uniques = algos.unique(level_codes)
				if has_na:
					na_idx = np.where(uniques == -1)[0]
					# Just ensure that -1 is in first position:
					uniques[[0, na_idx[0]]] = uniques[[na_idx[0], 0]]

				# codes get mapped from uniques to 0:len(uniques)
				# -1 (if present) is mapped to last position
				code_mapping = np.zeros(len(lev) + has_na)
				# ... and reassigned value -1:
				code_mapping[uniques] = np.arange(len(uniques)) - has_na

				level_codes = code_mapping[level_codes]

				# new levels are simple
				lev = lev.take(uniques[has_na:])

		new_levels.append(lev)
		new_codes.append(level_codes)

	result = multiindex.view()

	if changed:
		result._reset_identity()
		result._set_levels(new_levels, validate=False)
		result._set_codes(new_codes, validate=False)

	return result


def replace_levels(multiindex, level, new_label_array):
	"""
	Replace the labels in a level with a new order.

	Parameters
	----------
	multiindex : pandas.MultiIndex
	level : int
	new_label_array : array-like

	Returns
	-------
	pandas.MultiIndex
	"""

	levels = [i for i in multiindex.levels]
	codes = [i for i in multiindex.codes]

	new_label_array = np.asarray(new_label_array)
	old_label_array = multiindex.levels[level]
	where_in_new_label_array = {}
	for n, j in enumerate(old_label_array):
		try:
			where_in_new_label_array[n] = int(np.where(j == new_label_array)[0])
		except TypeError:
			raise ValueError(f"missing {j} in new_label_array, all the old labels must appear")

	old_codes = list(multiindex.codes[level])
	codes[level] = list(map(where_in_new_label_array.get, old_codes))
	levels[level] = new_label_array

	return pd.MultiIndex(levels=levels, codes=codes, names=multiindex.names)



import numpy, pandas
from collections import OrderedDict
from ..warning import ignore_warnings

import logging
from ..log import logger_name
logger = logging.getLogger(logger_name)

try:
	from ipywidgets import FloatProgress as ipywidgets_FloatProgress
except ImportError:
	ipywidgets_FloatProgress = ()

def debug_array(name, arr):
	print(f"<{name} shape={arr.shape}>")
	print(arr)
	print(f"</{name}>")


def _approx_fprime_helper(xk, f, epsilon, args=(), f0=None, *, status_widget=None):
	"""
	See ``approx_fprime``.  An optional initial function value arg is added.

	"""
	if f0 is None:
		f0_ = f(*((xk,) + args))
		try:
			f0 = f0_.copy()
		except AttributeError:
			f0 = f0_
	numpy.nan_to_num(f0, copy=False)
	try:
		f_shape = f0.shape
	except AttributeError:
		f_shape = ()
	grad = numpy.zeros((len(xk),)+f_shape, float)
	ei = numpy.zeros((len(xk),), float)
	for k in range(len(xk)):
		ei[k] = 1.0
		d = epsilon * ei
		grad[k] = ( numpy.nan_to_num(f(*((xk + d,) + args))) - f0) / d[k]
		ei[k] = 0.0
		if status_widget:
			status_widget("{} / {}".format(k, len(xk)))
	return grad


def _approx_fprime_helper_trailing(xk, f, epsilon, args=(), f0=None):
	"""
	See ``approx_fprime``.  An optional initial function value arg is added.

	"""
	if f0 is None:
		f0_ = f(*((xk,) + args))
		try:
			f0 = f0_.copy()
		except AttributeError:
			f0 = f0_
	numpy.nan_to_num(f0, copy=False)
	try:
		f_shape = f0.shape
	except AttributeError:
		f_shape = ()
	grad = numpy.zeros(f_shape+(len(xk),), float)
	ei = numpy.zeros((len(xk),), float)
	for k in range(len(xk)):
		ei[k] = 1.0
		d = epsilon * ei
		grad[...,k] = ( numpy.nan_to_num(f(*((xk + d,) + args))) - f0) / d[k]
		ei[k] = 0.0
	return grad


def approx_fprime(xk, f, epsilon=None, trailing=False, *args, status_widget=None):
	"""Finite-difference approximation of the gradient of a scalar function.

	Parameters
	----------
	xk : array_like
	    The coordinate vector at which to determine the gradient of `f`.
	f : callable
	    The function of which to determine the gradient (partial derivatives).
	    Should take `xk` as first argument, other arguments to `f` can be
	    supplied in ``*args``.  Should return a scalar, the value of the
	    function at `xk`.
	epsilon : array_like
	    Increment to `xk` to use for determining the function gradient.
	    If a scalar, uses the same finite difference delta for all partial
	    derivatives.  If an array, should contain one value per element of
	    `xk`.
	\\*args : args, optional
	    Any other arguments that are to be passed to `f`.

	Returns
	-------
	grad : ndarray
	    The partial derivatives of `f` to `xk`.

	See Also
	--------
	check_grad : Check correctness of gradient function against approx_fprime.

	Notes
	-----
	The function gradient is determined by the forward finite difference
	formula::

	             f(xk[i] + epsilon[i]) - f(xk[i])
	    f'[i] = ---------------------------------
	                        epsilon[i]

	The main use of `approx_fprime` is in scalar function optimizers like
	`fmin_bfgs`, to determine numerically the Jacobian of a function.

	"""
	if epsilon is None:
		epsilon = numpy.sqrt(numpy.finfo(float).eps)
	if trailing:
		return _approx_fprime_helper_trailing(xk, f, epsilon, args=args)
	else:
		return _approx_fprime_helper(xk, f, epsilon, args=args, status_widget=status_widget)

def similarity(a,b, to_zero=None):
	"""
	The similarity between two values or arrays.

	Returns a similarity measure indicating approximately the number of
	significant figures in common between the two values.  Returns 100
	if the values are exactly equal.

	Parameters
	----------
	a,b : numeric or arrays of same size
		Scalar values are converted to numpy arrays.
	to_zero : numeric, optional
		A threshold level for values that can be
		considered to match zero, where if one value is
		exactly zero and the other is not, the similarity
		is equal to the inverse absolute value of the other value
		multiplied by `to_zero`

	Returns
	-------
	ndarray

	"""
	a = numpy.asarray(a)
	b = numpy.asarray(b)
	magnitude = numpy.fmax(numpy.fabs(a),numpy.fabs(b))
	difference = numpy.abs(a-b)
	with numpy.errstate(divide='ignore',invalid='ignore'):
		similar = -numpy.log10(difference/magnitude)
	similar[a==b] = 100
	if to_zero:
		with ignore_warnings():
			similar[a == 0] = (to_zero / difference)[a == 0]
			similar[b == 0] = (to_zero / difference)[b == 0]
	return similar


def _color_poor_similarity(val):
	"""
	Takes a scalar similarity and returns a string with
	the css property `'color: red'` for negative
	strings, black otherwise.
	"""
	if val < 3:
		return 'color: #FF0000'
	if val < 4.5:
		return 'color: #BB0000'
	if val < 6:
		return 'color: #880000'
	return 'color: black'



def check_gradient(func, grad, x0, epsilon=None, names=None, require_sim=None, stylize=True, skip_zeros=True, trailing=False):
	"""Check the gradient.

	Check the correctness of a gradient function by comparing it against a
	(forward) finite-difference approximation of the gradient.

	Parameters
	----------
	func : callable ``func(x0, *args)``
	    Function whose derivative is to be checked.
	grad : callable ``grad(x0, *args)``
	    Gradient of `func`.
	x0 : ndarray
	    Points to check `grad` against forward difference approximation of grad
	    using `func`.
	args : \\*args, optional
	    Extra arguments passed to `func` and `grad`.
	epsilon : float, optional
	    Step size used for the finite difference approximation. It defaults to
	    ``sqrt(numpy.finfo(float).eps)``, which is approximately 1.49e-08.
	names : list or pandas.Index
		A list of names by position in the returned gradient.
	require_sim : float, optional
		If any item has similarity less than this, a GradientCheckError is raise

	Returns
	-------
	err : float
	    The square root of the sum of squares (i.e. the 2-norm) of the
	    difference between ``grad(x0, *args)`` and the finite difference
	    approximation of `grad` using func at the points `x0`.
	"""
	if epsilon is None:
		epsilon = numpy.sqrt(numpy.finfo(float).eps)
	g_f = approx_fprime(x0, func, epsilon, trailing=trailing)
	g_a = grad(x0)

	g_f_ = g_f.reshape(-1)
	if isinstance(g_a, pandas.Series):
		g_a = g_a.values
	g_a_ = g_a.reshape(-1)
	lister = [
			# ('x', x0),
			('finite_diff', g_f_),
			('analytic', g_a_),
			('diff', g_f_ - g_a_),
			('similarity', similarity(g_f_, g_a_, to_zero=0.01)),
		]
	try:
		f = pandas.DataFrame.from_dict(OrderedDict(lister))
	except:
		print(lister)
		raise
	if names is not None and len(names)==g_a_.size:
		f.index = names
	elif names is not None and len(names)**2 == g_a_.size:
		from itertools import product
		f.index = ["{}, {}".format(n1,n2) for n1,n2 in product(names, repeat=2)]
	elif names is not None:
		try:
			f.index = numpy.broadcast_to(names, g_f.shape).reshape(-1)
		except ValueError:
			f.index = numpy.broadcast_to(names, g_f.T.shape).T.reshape(-1)

	if skip_zeros:
		f = f[(f['finite_diff']!=0) | (f['analytic']!=0)]

	if not stylize:
		return f
	try:
		return f.style.applymap(_color_poor_similarity, subset='similarity')
	except ValueError:
		return f


def assert_gradient_check(*args, min_similarity=4, **kwargs):
	kwargs['stylize'] = False
	df = check_gradient(*args, **kwargs)
	assert df['similarity'].min() >= min_similarity


import larch
import larch.model.linear_math
from larch.model.linear import ParameterRef_C, DataRef_C, LinearComponent_C, LinearFunction_C, _null_
from larch import P, X, PX
import keyword
import pytest

def test_parameter_c():

	p = ParameterRef_C("hsh")
	assert "hsh" == p
	assert p == "hsh"
	assert not keyword.iskeyword(p)
	assert hash(p) == hash("hsh")
	assert repr(p) == "P.hsh"
	assert p == P.hsh
	assert p == P("hsh")
	assert p == P['hsh']

def test_data_c():

	d = DataRef_C("hsh")
	assert "hsh" == d
	assert d == "hsh"
	assert not keyword.iskeyword(d)
	assert hash(d) == hash("hsh")
	assert repr(d) == "X.hsh"

	assert d == X.hsh
	assert d == X("hsh")
	assert d == X['hsh']

	p = ParameterRef_C("hsh")
	assert not p == d
	assert p != d

def test_data_c_math():

	assert X.Aaa + X.Bbb == X("Aaa+Bbb")
	assert X.Aaa - X.Bbb == X("Aaa-Bbb")
	assert X.Aaa * X.Bbb == X("Aaa*Bbb")
	assert X.Aaa / X.Bbb == X("Aaa/Bbb")
	assert X.Aaa & X.Bbb == X("Aaa&Bbb")
	assert X.Aaa | X.Bbb == X("Aaa|Bbb")
	assert X.Aaa ^ X.Bbb == X("Aaa^Bbb")
	assert X.Aaa ** X.Bbb == X("Aaa**Bbb")
	assert X.Zzz / X.Aaa + X.Vvv * X.Bbb == X('(Zzz/Aaa)+(Vvv*Bbb)')
	assert +X.Aaa == X("Aaa")
	assert -X.Aaa == X("-Aaa")

	assert X.Aaa + 2 == X("Aaa+2")
	assert X.Aaa - 2 == X("Aaa-2")
	assert X.Aaa * 2 == LinearComponent_C(param=_null_, scale=2, data='Aaa')
	assert X.Aaa / 2 == X("Aaa/2")
	assert X.Aaa & 2 == X("Aaa&2")
	assert X.Aaa | 2 == X("Aaa|2")
	assert X.Aaa ^ 2 == X("Aaa^2")
	assert X.Aaa ** 2 == X("Aaa**2")

	assert 2 + X.Aaa == X("2+Aaa")
	assert 2 - X.Aaa == X("2-Aaa")
	assert 2 * X.Aaa == LinearComponent_C(param=_null_, scale=2, data='Aaa')
	assert 2 / X.Aaa == X("2/Aaa")
	assert 2 & X.Aaa == X("2&Aaa")
	assert 2 | X.Aaa == X("2|Aaa")
	assert 2 ^ X.Aaa == X("2^Aaa")
	assert 2 ** X.Aaa == X("2**Aaa")

	assert X.Aaa + 0 == X.Aaa
	assert 0 + X.Aaa == X.Aaa

	assert X.Aaa * 1 == X.Aaa
	assert 1 * X.Aaa == X.Aaa

	with pytest.raises(TypeError):
		_ = X.Aaa + "Plain String"

	with pytest.raises(TypeError):
		_ = X.Aaa - "Plain String"

	with pytest.raises(TypeError):
		_ = X.Aaa * "Plain String"

	with pytest.raises(TypeError):
		_ = X.Aaa / "Plain String"


def test_ref_gen():

	assert X["Asd"] == X("Asd") == X.Asd
	assert P["Asd"] == P("Asd") == P.Asd
	assert X.Asd != P.Asd

def test_linear_func():

	assert LinearComponent_C(param="pname", data="dname") == P.pname * X.dname

	assert type(list(P.singleton + P.pname * X.dname)[0]) is LinearComponent_C
	assert type(list(P.singleton + P.pname * X.dname)[1]) is LinearComponent_C

	assert type(list( + P.pname * X.dname + P.singleton)[0]) is LinearComponent_C
	assert type(list( + P.pname * X.dname + P.singleton)[1]) is LinearComponent_C



	assert list(-(P.pname * X.dname + P.singleton)) == [
		LinearComponent_C('pname', 'dname', -1.0),
		LinearComponent_C('singleton', '1', -1.0),
	]
	assert list(-(P.pname * X.dname - P.singleton)) == [
		LinearComponent_C('pname', 'dname', -1.0),
		LinearComponent_C('singleton', '1', 1.0),
	]

	assert list((P.pname * X.dname - P.singleton) * X.Sss) == [
		LinearComponent_C(param='pname', data='dname*Sss', scale=1.0),
		LinearComponent_C(param='singleton', data='Sss', scale=-1.0),
	]

	assert list(sum(PX(i) for i in ['Aaa', 'Bbb'])) == [
		LinearComponent_C(param='Aaa', data='Aaa', scale=1.0),
		LinearComponent_C(param='Bbb', data='Bbb', scale=1.0),
	]


	u = P.Aaa * X.Aaa + P.Bbb * X.Bbb
	u += P.Ccc * X.Ccc

	assert u == P.Aaa * X.Aaa + P.Bbb * X.Bbb + P.Ccc * X.Ccc

	assert P.ppp * X.xxx * 1.234 == P.ppp * 1.234 * X.xxx
	assert P.ppp * X.xxx * 1.234 == X.xxx * P.ppp * 1.234
	assert P.ppp * X.xxx * 1.234 == X.xxx * 1.234 * P.ppp
	assert P.ppp * X.xxx * 1.234 == 1.234 * X.xxx * P.ppp
	assert P.ppp * X.xxx * 1.234 == 1.234 * P.ppp * X.xxx

	assert (P.ppp * X.xxx) * 1.234 == P.ppp * (1.234 * X.xxx)
	assert (P.ppp * X.xxx) * 1.234 == X.xxx * (P.ppp * 1.234)
	assert (P.ppp * X.xxx) * 1.234 == X.xxx * (1.234 * P.ppp)
	assert (P.ppp * X.xxx) * 1.234 == 1.234 * (X.xxx * P.ppp)
	assert (P.ppp * X.xxx) * 1.234 == 1.234 * (P.ppp * X.xxx)

	assert (P.ppp * X.xxx) * 1.234 == (P.ppp * 1.234) * X.xxx
	assert (P.ppp * X.xxx) * 1.234 == (X.xxx * P.ppp) * 1.234
	assert (P.ppp * X.xxx) * 1.234 == (X.xxx * 1.234) * P.ppp
	assert (P.ppp * X.xxx) * 1.234 == (1.234 * X.xxx) * P.ppp
	assert (P.ppp * X.xxx) * 1.234 == (1.234 * P.ppp) * X.xxx

	assert (P.ppp * X.xxx * 1.234) == P.ppp * (1.234 * X.xxx)
	assert (P.ppp * X.xxx * 1.234) == X.xxx * (P.ppp * 1.234)
	assert (P.ppp * X.xxx * 1.234) == X.xxx * (1.234 * P.ppp)
	assert (P.ppp * X.xxx * 1.234) == 1.234 * (X.xxx * P.ppp)
	assert (P.ppp * X.xxx * 1.234) == 1.234 * (P.ppp * X.xxx)

	assert (P.ppp * X.xxx * 1.234) == (P.ppp * 1.234) * X.xxx
	assert (P.ppp * X.xxx * 1.234) == (X.xxx * P.ppp) * 1.234
	assert (P.ppp * X.xxx * 1.234) == (X.xxx * 1.234) * P.ppp
	assert (P.ppp * X.xxx * 1.234) == (1.234 * X.xxx) * P.ppp
	assert (P.ppp * X.xxx * 1.234) == (1.234 * P.ppp) * X.xxx

	assert P.ppp * (X.xxx * 1.234) == P.ppp * (1.234 * X.xxx)
	assert P.ppp * (X.xxx * 1.234) == X.xxx * (P.ppp * 1.234)
	assert P.ppp * (X.xxx * 1.234) == X.xxx * (1.234 * P.ppp)
	assert P.ppp * (X.xxx * 1.234) == 1.234 * (X.xxx * P.ppp)
	assert P.ppp * (X.xxx * 1.234) == 1.234 * (P.ppp * X.xxx)

	assert P.ppp * (X.xxx * 1.234) == (P.ppp * 1.234) * X.xxx
	assert P.ppp * (X.xxx * 1.234) == (X.xxx * P.ppp) * 1.234
	assert P.ppp * (X.xxx * 1.234) == (X.xxx * 1.234) * P.ppp
	assert P.ppp * (X.xxx * 1.234) == (1.234 * X.xxx) * P.ppp
	assert P.ppp * (X.xxx * 1.234) == (1.234 * P.ppp) * X.xxx

	assert (P.ppp * X.xxx) * X.xxx == P.ppp * X('xxx*xxx')
	assert (P.ppp * X.xxx) * (P("_") * X.xxx) == P.ppp * X('xxx*xxx')
	assert (P("_") * X.xxx) * (P.ppp * X.xxx) == P.ppp * X('xxx*xxx')

	# Test squaring a boolean
	assert (P.ppp * X('boolean(xxx)')) * X('boolean(xxx)') == P.ppp * X('boolean(xxx)')
	assert (P.ppp * X('boolean(xxx)')) * (P("_") * X('boolean(xxx)')) == P.ppp * X('boolean(xxx)')

	assert ((P.p1 * X.x1 + P.p2 * X.x2) * (P('_') * 1.1 * X.x1 + P('_') * 2 * X.x2)) == (
			P.p1 * 1.1 * X('x1*x1') + P.p1 * 2.0 * X('x1*x2') + P.p2 * 1.1 * X('x2*x1') + P.p2 * 2.0 * X('x2*x2')
	)

def test_piecewise_linear():
	from larch.util.data_expansion import piecewise_linear

	func = piecewise_linear(X.DataName, P.ParamName, [3, 5, 7])
	assert func[0] == P('ParamName ① up to 3') * X('piece(DataName,None,3)')
	assert func[1] == P('ParamName ② 3 to 5') * X('piece(DataName,3,5)')
	assert func[2] == P('ParamName ③ 5 to 7') * X('piece(DataName,5,7)')
	assert func[3] == P('ParamName ④ over 7') * X('piece(DataName,7,None)')
	assert len(func) == 4

	func = piecewise_linear(X.DataName, breaks=[3, 5, 7])
	assert func[0] == P('DataName ① up to 3') * X('piece(DataName,None,3)')
	assert func[1] == P('DataName ② 3 to 5') * X('piece(DataName,3,5)')
	assert func[2] == P('DataName ③ 5 to 7') * X('piece(DataName,5,7)')
	assert func[3] == P('DataName ④ over 7') * X('piece(DataName,7,None)')
	assert len(func) == 4

	func = piecewise_linear('GenName', breaks=[3, 5, 7])
	assert func[0] == P('GenName ① up to 3') * X('piece(GenName,None,3)')
	assert func[1] == P('GenName ② 3 to 5') * X('piece(GenName,3,5)')
	assert func[2] == P('GenName ③ 5 to 7') * X('piece(GenName,5,7)')
	assert func[3] == P('GenName ④ over 7') * X('piece(GenName,7,None)')
	assert len(func) == 4

	with pytest.raises(ValueError):
		func = piecewise_linear('GenName', [3, 5, 7])


def test_piecewise_expansion():
	import pandas, io, numpy
	from larch.util.data_expansion import piecewise_expansion, piecewise_linear

	df = pandas.DataFrame(
		numpy.linspace(0, 10, 25),
		columns=['DataName'],
	)

	expanded = piecewise_expansion(df, [3, 5, 7])

	s = '''
	    piece(DataName,None,3)  piece(DataName,3,5)  piece(DataName,5,7)  piece(DataName,7,None)
	0                 0.000000             0.000000             0.000000                0.000000
	1                 0.416667             0.000000             0.000000                0.000000
	2                 0.833333             0.000000             0.000000                0.000000
	3                 1.250000             0.000000             0.000000                0.000000
	4                 1.666667             0.000000             0.000000                0.000000
	5                 2.083333             0.000000             0.000000                0.000000
	6                 2.500000             0.000000             0.000000                0.000000
	7                 2.916667             0.000000             0.000000                0.000000
	8                 3.000000             0.333333             0.000000                0.000000
	9                 3.000000             0.750000             0.000000                0.000000
	10                3.000000             1.166667             0.000000                0.000000
	11                3.000000             1.583333             0.000000                0.000000
	12                3.000000             2.000000             0.000000                0.000000
	13                3.000000             2.000000             0.416667                0.000000
	14                3.000000             2.000000             0.833333                0.000000
	15                3.000000             2.000000             1.250000                0.000000
	16                3.000000             2.000000             1.666667                0.000000
	17                3.000000             2.000000             2.000000                0.083333
	18                3.000000             2.000000             2.000000                0.500000
	19                3.000000             2.000000             2.000000                0.916667
	20                3.000000             2.000000             2.000000                1.333333
	21                3.000000             2.000000             2.000000                1.750000
	22                3.000000             2.000000             2.000000                2.166667
	23                3.000000             2.000000             2.000000                2.583333
	24                3.000000             2.000000             2.000000                3.000000
	'''

	correct = pandas.read_csv(io.StringIO(s), sep='\s+')

	pandas.testing.assert_frame_equal(expanded, correct)

	func = piecewise_linear(X.DataName, P.ParamName, [3, 5, 7])
	expanded2 = piecewise_expansion(df, func)
	pandas.testing.assert_frame_equal(expanded2, correct)


def test_pmath():
	m = larch.Model()
	m.utility_ca = P.Aaa * X.Aaa + P.Bbb * X.Bbb + P.Ccc
	m.set_values(Aaa=12, Bbb=20, Ccc=2)

	assert P.Aaa.value(m) == 12
	assert P.Bbb.value(m) == 20
	assert P.Ccc.value(m) == 2

	with pytest.raises(KeyError):
		P.Ddd.value(m)

	assert P.Ddd.value(m, {'Ddd': 123}) == 123

	y = P.Aaa + P.Bbb
	assert y.value(m) == 12 + 20
	assert repr(y) == 'P.Aaa + P.Bbb'
	assert repr(y + y) == 'P.Aaa + P.Bbb + P.Aaa + P.Bbb'
	assert isinstance(y, larch.model.linear.LinearFunction_C)
	assert isinstance(y + y, larch.model.linear.LinearFunction_C)

	y = P.Aaa / P.Bbb
	assert y.value(m) == 12 / 20
	assert y.string(m) == "0.6"
	assert repr(y) == 'P.Aaa / P.Bbb'
	assert repr(y + y) == 'P.Aaa / P.Bbb + P.Aaa / P.Bbb'

	y = P.Aaa * 60 / (P.Bbb * 100)
	y.set_fmt("${:0.2f}/hr")
	assert y.value(m) == 12 * 60 / (20 * 100)
	assert y.string(m) == '$0.36/hr'

	y = P.Aaa + P.Bbb / P.Ccc
	assert y.value(m) == 12 + 20 / 2
	assert isinstance(y, larch.model.linear_math.ParameterAdd)
	assert repr(y) == 'P.Aaa + P.Bbb / P.Ccc'

	y = (P.Aaa + P.Bbb) + P.Ccc * P.Aaa
	assert isinstance(y, larch.model.linear_math.ParameterAdd)
	assert y.value(m) == (12 + 20) + 2 * 12

	y = (P.Aaa + P.Bbb) + P.Ccc / P.Aaa
	assert isinstance(y, larch.model.linear_math.ParameterAdd)
	assert y.value(m) == (12 + 20) + 2 / 12

	y = (P.Aaa + P.Bbb + P.Ccc) * P.Aaa
	assert isinstance(y, larch.model.linear_math.ParameterMultiply)
	assert y.value(m) == (12 + 20 + 2) * 12

	y = (P.Aaa + P.Bbb + P.Ccc) / P.Aaa
	assert isinstance(y, larch.model.linear_math.ParameterDivide)
	assert y.value(m) == (12 + 20 + 2) / 12
	assert repr(y) == '(P.Aaa + P.Bbb + P.Ccc) / P.Aaa'

	y = (P.Aaa + P.Bbb + P.Ccc) - P.Aaa
	assert isinstance(y, larch.model.linear.LinearFunction_C)
	assert y.value(m) == 12 + 20 + 2 - 12


def test_pvalue():
	m = larch.Model()
	m.utility_ca = P.Aaa * X.Aaa + P.Bbb * X.Bbb + P.Ccc
	m.set_values(Aaa=12, Bbb=20, Ccc=2)

	assert P.Aaa.value(m) == 12
	assert P.Bbb.value(m) == 20
	assert P.Ccc.value(m) == 2

	with pytest.raises(KeyError):
		P.Ddd.value(m)

	assert P.Ddd.value(m, {'Ddd': 123}) == 123

	y = P.Ddd / P.Bbb
	with pytest.raises(KeyError):
		m.pvalue(y, log_errors=False)
	assert m.pformat(y) == 'NA'

	y = P.Aaa + P.Bbb
	assert m.pvalue(y) == 12 + 20
	assert m.pformat(y) == str(12 + 20)

	y = P.Aaa / P.Bbb
	assert m.pvalue(y) == 12 / 20
	assert m.pformat(y) == "0.6"

	y = P.Aaa * 60 / (P.Bbb * 100)
	y.set_fmt("${:0.2f}/hr")
	assert m.pvalue(y) == 12 * 60 / (20 * 100)
	assert m.pformat(y) == '$0.36/hr'

	y = P.Aaa + P.Bbb / P.Ccc
	assert m.pvalue(y) == 12 + 20 / 2
	assert m.pformat(y) == '22'

	y = (P.Aaa + P.Bbb) + P.Ccc * P.Aaa
	assert m.pvalue(y) == (12 + 20) + 2 * 12
	assert m.pformat(y) == "{:.3g}".format((12 + 20) + 2 * 12)

	y = (P.Aaa + P.Bbb) + P.Ccc / P.Aaa
	assert m.pvalue(y) == (12 + 20) + 2 / 12
	assert m.pformat(y) == "{:.3g}".format((12 + 20) + 2 / 12)

	y = (P.Aaa + P.Bbb + P.Ccc) * P.Aaa
	assert m.pvalue(y) == (12 + 20 + 2) * 12
	assert m.pformat(y) == "{:.3g}".format((12 + 20 + 2) * 12)

	y = (P.Aaa + P.Bbb + P.Ccc) / P.Aaa
	assert m.pvalue(y) == (12 + 20 + 2) / 12
	assert m.pformat(y) == "{:.3g}".format((12 + 20 + 2) / 12)

	y = (P.Aaa + P.Bbb + P.Ccc) - P.Aaa
	assert m.pvalue(y) == 12 + 20 + 2 - 12
	assert m.pformat(y) == "{:.3g}".format(12 + 20 + 2 - 12)

	y = (P.Aaa + P.Bbb * 2) + P.Ccc * P.Aaa
	assert m.pvalue(y) == (12 + 20 * 2) + 2 * 12
	assert m.pformat(y) == "{:.3g}".format((12 + 20 * 2) + 2 * 12)

	y = (P.Aaa + P.Bbb) + P.Ccc / P.Aaa * 3
	assert m.pvalue(y) == (12 + 20) + 2 / 12 * 3
	assert m.pformat(y) == "{:.3g}".format((12 + 20) + 2 / 12 * 3)

	y = (P.Aaa + P.Bbb + P.Ccc) * P.Aaa * 3
	assert m.pvalue(y) == (12 + 20 + 2) * 12 * 3
	assert m.pformat(y) == "{:.3g}".format((12 + 20 + 2) * 12 * 3)

	y = (P.Aaa + P.Bbb + P.Ccc) / P.Aaa * 3
	assert m.pvalue(y) == (12 + 20 + 2) / 12 * 3
	assert m.pformat(y) == "{:.3g}".format((12 + 20 + 2) / 12 * 3)

	y = (P.Aaa + P.Bbb + P.Ccc) - P.Aaa * 3
	assert m.pvalue(y) == 12 + 20 + 2 - 12 * 3
	assert m.pformat(y) == "{:.3g}".format(12 + 20 + 2 - 12 * 3)

	y = (11 + P.Aaa + P.Bbb * 2) + P.Ccc * P.Aaa
	assert m.pvalue(y) == (11 + 12 + 20 * 2) + 2 * 12
	assert m.pformat(y) == "{:.3g}".format((11 + 12 + 20 * 2) + 2 * 12)

	y = (11 + P.Aaa + P.Bbb) + P.Ccc / P.Aaa * 3
	assert m.pvalue(y) == (11 + 12 + 20) + 2 / 12 * 3
	assert m.pformat(y) == "{:.3g}".format((11 + 12 + 20) + 2 / 12 * 3)

	y = (11 + P.Aaa + P.Bbb + P.Ccc) * P.Aaa * 3
	assert m.pvalue(y) == (11 + 12 + 20 + 2) * 12 * 3
	assert m.pformat(y) == "{:.3g}".format((11 + 12 + 20 + 2) * 12 * 3)

	y = (11 + P.Aaa + P.Bbb + P.Ccc) / P.Aaa * 3
	assert m.pvalue(y) == (11 + 12 + 20 + 2) / 12 * 3
	assert m.pformat(y) == "{:.3g}".format((11 + 12 + 20 + 2) / 12 * 3)

	y = (11 + P.Aaa + P.Bbb + P.Ccc) - P.Aaa * 3
	assert m.pvalue(y) == 11 + 12 + 20 + 2 - 12 * 3
	assert m.pformat(y) == "{:.3g}".format(11 + 12 + 20 + 2 - 12 * 3)

	with pytest.raises(NotImplementedError):
		P.Aaa * 10 + P.Bbb / 10 * X.Yyy + 2

	y = P.Aaa * 10 + P.Bbb / 10 + 2
	assert m.pvalue(y) == 12 * 10 + 20 / 10 + 2
	assert m.pformat(y) == "{:.3g}".format(12 * 10 + 20 / 10 + 2)

	with pytest.raises(Exception):
		m.utility_ca = P.Aaa + P.Ccc + 5

	yd = {
		'Yxx': P.Aaa + P.Bbb / P.Ccc,
		'Yyy': P.Aaa + P.Bbb,
		'Yzz': P.Aaa * 60 / (P.Bbb * 100),
	}
	assert m.pvalue(yd) == {
		'Yxx': 12 + 20 / 2,
		'Yyy': 12 + 20,
		'Yzz': 12 * 60 / (20 * 100),
	}
	assert m.pformat(yd) == {
		'Yxx': "{:.3g}".format(12 + 20 / 2),
		'Yyy': "{:.3g}".format(12 + 20),
		'Yzz': "{:.3g}".format(12 * 60 / (20 * 100)),
	}


def test_pmath_in_utility():
	d = larch.examples.MTC()
	m0 = larch.Model(dataservice=d)

	m0.utility_co[2] = P("ASC_SR2") * 10 + P("hhinc#2") / 10 * X("hhinc")
	m0.utility_co[3] = P("ASC_SR3P") * 10 + P("hhinc#3") / 10 * X("hhinc")
	m0.utility_co[4] = P("ASC_TRAN") * 10 + P("hhinc#4") / 10 * X("hhinc")
	m0.utility_co[5] = P("ASC_BIKE") * 10 + P("hhinc#5") / 10 * X("hhinc")
	m0.utility_co[6] = P("ASC_WALK") * 10 + P("hhinc#6") / 10 * X("hhinc")

	m0.utility_ca = (
			+ P("nonmotorized_time") / 10. * X("(altnum>4) * tottime")
			+ P("motorized_ovtt") * 10 * X("(altnum <= 4) * ovtt")
			+ P("motorized_ivtt") * X("(altnum <= 4) * ivtt")
			+ PX("totcost")
	)
	m0.availability_var = '_avail_'
	m0.choice_ca_var = '_choice_'

	m1 = larch.Model(dataservice=d)

	m1.utility_co[2] = P("ASC_SR2") * X('10') + P("hhinc#2") * X("hhinc/10")
	m1.utility_co[3] = P("ASC_SR3P") * X('10') + P("hhinc#3") * X("hhinc/10")
	m1.utility_co[4] = P("ASC_TRAN") * X('10') + P("hhinc#4") * X("hhinc/10")
	m1.utility_co[5] = P("ASC_BIKE") * X('10') + P("hhinc#5") * X("hhinc/10")
	m1.utility_co[6] = P("ASC_WALK") * X('10') + P("hhinc#6") * X("hhinc/10")

	m1.utility_ca = (
			+ P("nonmotorized_time") * X("(altnum>4) * tottime / 10")
			+ P("motorized_ovtt") * X("(altnum <= 4) * ovtt * 10")
			+ P("motorized_ivtt") * X("(altnum <= 4) * ivtt")
			+ PX("totcost")
	)
	m1.availability_var = '_avail_'
	m1.choice_ca_var = '_choice_'

	m0.load_data()
	m1.load_data()

	r0 = m0.maximize_loglike(quiet=True)
	r1 = m1.maximize_loglike(quiet=True)
	assert r0.loglike == pytest.approx(-3587.6430040944942)
	assert r1.loglike == pytest.approx(-3587.6430040944942)

	m0.calculate_parameter_covariance()
	m1.calculate_parameter_covariance()
	t = {
		'ASC_BIKE': -5.318650574990901,
		'ASC_SR2': -22.291563439182628,
		'ASC_SR3P': -22.174552606750527,
		'ASC_TRAN': -3.293923857045225,
		'ASC_WALK': 1.6172450189610719,
		'hhinc#2': -1.4000897138949544,
		'hhinc#3': 0.12900984170888324,
		'hhinc#4': -3.0601742475362923,
		'hhinc#5': -2.333410249527477,
		'hhinc#6': -3.048442130390144,
		'motorized_ivtt': -0.4116740527068954,
		'motorized_ovtt': -12.958446214791113,
		'nonmotorized_time': -11.789244777056298,
		'totcost': -20.19350165272386,
	}
	assert dict(m0.pf['t_stat']) == pytest.approx(t, rel=1e-5)
	assert dict(m1.pf['t_stat']) == pytest.approx(t, rel=1e-5)

	assert (m0.get_value(P.motorized_ivtt) * 60) / (m0.get_value(P.totcost) * 100) == pytest.approx(0.3191492801963062)
	assert m0.get_value( (P.motorized_ivtt * 60) / (P.totcost * 100) ) == pytest.approx(0.3191492801963062)
	assert (m1.get_value(P.motorized_ivtt) * 60) / (m1.get_value(P.totcost) * 100) == pytest.approx(0.3191492801963062)
	assert m1.get_value( (P.motorized_ivtt * 60) / (P.totcost * 100) ) == pytest.approx(0.3191492801963062)

def test_linear_function_iadd():
	# Test inplace add on unattached LinearFunction_C
	lf = P.tottime * X.tottime + P.totcost * X.totcost
	lf += X("totcost*tottime") * P("fake")
	assert lf == P.tottime * X.tottime + P.totcost * X.totcost + X("totcost*tottime") * P("fake")
	# Test inplace add on attached LinearFunction_C
	m = larch.Model(utility_ca=P.tottime * X.tottime + P.totcost * X.totcost)
	m.utility_ca += X("totcost*tottime") * P("fake")
	xx = P.tottime * X.tottime + P.totcost * X.totcost + X("totcost*tottime") * P("fake")
	assert m.utility_ca == xx
